from __future__ import annotations

import itertools
from typing import TYPE_CHECKING

import pytest

from qcarchivetesting import load_molecule_data
from qcportal.record_models import PriorityEnum, RecordStatusEnum
from qcportal.singlepoint import QCSpecification, SinglepointDriver
from qcportal.utils import now_at_utc
from .testing_helpers import submit_procedure_data, run_procedure_data, compare_singlepoint_specs, test_specs

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from qcportal import PortalClient


def test_singlepoint_client_tag_priority(snowflake_client: PortalClient):
    water = load_molecule_data("water_dimer_minima")

    for tag, priority in itertools.product(["*", "tag99"], list(PriorityEnum)):
        meta1, id1 = snowflake_client.add_singlepoints(
            [water],
            "prog",
            SinglepointDriver.energy,
            "hf",
            "sto-3g",
            {"tag_priority": [tag, priority]},
            None,
            compute_priority=priority,
            compute_tag=tag,
        )
        assert meta1.n_inserted == 1

        rec = snowflake_client.get_records(id1, include=["task"])
        assert rec[0].task.compute_tag == tag
        assert rec[0].task.compute_priority == priority


@pytest.mark.parametrize("spec", test_specs)
def test_singlepoint_client_add_get(submitter_client: PortalClient, spec: QCSpecification):
    water = load_molecule_data("water_dimer_minima")
    hooh = load_molecule_data("hooh")
    ne4 = load_molecule_data("neon_tetramer")
    all_mols = [water, hooh, ne4]

    time_0 = now_at_utc()
    meta, id = submitter_client.add_singlepoints(
        all_mols,
        spec.program,
        spec.driver,
        spec.method,
        spec.basis,
        spec.keywords,
        spec.protocols,
        "tag1",
        PriorityEnum.high,
    )
    time_1 = now_at_utc()

    recs = submitter_client.get_singlepoints(id, include=["task", "molecule"])

    for r in recs:
        assert r.record_type == "singlepoint"
        assert r.record_type == "singlepoint"
        assert compare_singlepoint_specs(spec, r.specification)

        assert r.status == RecordStatusEnum.waiting
        assert r.children_status == {}

        assert r.task.function is None
        assert r.task.compute_tag == "tag1"
        assert r.task.compute_priority == PriorityEnum.high
        assert r.creator_user == submitter_client.username
        assert time_0 < r.created_on < time_1
        assert time_0 < r.modified_on < time_1

    assert recs[0].molecule == water
    assert recs[1].molecule == hooh
    assert recs[2].molecule == ne4


@pytest.mark.parametrize("spec", test_specs)
@pytest.mark.parametrize("find_existing", [True, False])
def test_singlepoint_client_add_duplicate(submitter_client: PortalClient, spec: QCSpecification, find_existing: bool):
    water = load_molecule_data("water_dimer_minima")
    hooh = load_molecule_data("hooh")
    ne4 = load_molecule_data("neon_tetramer")
    all_mols = [water, hooh, ne4]

    meta, id = submitter_client.add_singlepoints(
        all_mols,
        spec.program,
        spec.driver,
        spec.method,
        spec.basis,
        spec.keywords,
        spec.protocols,
        "tag1",
        PriorityEnum.high,
        find_existing=True,
    )

    assert meta.n_inserted == len(all_mols)

    meta, id2 = submitter_client.add_singlepoints(
        all_mols,
        spec.program,
        spec.driver,
        spec.method,
        spec.basis,
        spec.keywords,
        spec.protocols,
        "tag1",
        PriorityEnum.high,
        find_existing=find_existing,
    )

    if find_existing:
        assert meta.n_existing == len(all_mols)
        assert meta.n_inserted == 0
        assert id == id2
    else:
        assert meta.n_existing == 0
        assert meta.n_inserted == len(all_mols)
        assert set(id).isdisjoint(id2)


def test_singlepoint_client_add_existing_molecule(snowflake_client: PortalClient):
    spec = test_specs[0]

    water = load_molecule_data("water_dimer_minima")
    hooh = load_molecule_data("hooh")
    ne4 = load_molecule_data("neon_tetramer")
    all_mols = [water, hooh, ne4]

    # Add a molecule separately
    _, mol_ids = snowflake_client.add_molecules([ne4])

    # Now add records
    meta, ids = snowflake_client.add_singlepoints(
        all_mols,
        spec.program,
        spec.driver,
        spec.method,
        spec.basis,
        spec.keywords,
        spec.protocols,
        "tag1",
        PriorityEnum.high,
    )

    assert meta.success
    recs = snowflake_client.get_singlepoints(ids, include=["molecule"])

    assert len(recs) == 3
    assert recs[2].molecule_id == mol_ids[0]
    assert recs[2].molecule == ne4


def test_singlepoint_client_delete(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    sp_id = run_procedure_data(storage_socket, activated_manager_name, "sp_psi4_peroxide_energy_wfn")

    # deleting with children is ok (even though we don't have children)
    meta = snowflake_client.delete_records(sp_id, soft_delete=True, delete_children=True)
    assert meta.success
    assert meta.deleted_idx == [0]

    meta = snowflake_client.delete_records(sp_id, soft_delete=False, delete_children=True)
    assert meta.success
    assert meta.deleted_idx == [0]

    recs = snowflake_client.get_singlepoints(sp_id, missing_ok=True)
    assert recs is None


def test_singlepoint_client_query(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    id_1, _ = submit_procedure_data(storage_socket, "sp_psi4_benzene_energy_2")
    id_2, _ = submit_procedure_data(storage_socket, "sp_psi4_peroxide_energy_wfn")
    id_3, _ = submit_procedure_data(storage_socket, "sp_rdkit_water_energy")

    recs = snowflake_client.get_singlepoints([id_1, id_2, id_3])

    # query for molecule
    query_res = snowflake_client.query_singlepoints(molecule_id=[recs[1].molecule_id])
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    # query for program
    query_res = snowflake_client.query_singlepoints(program="psi4")
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    # query for basis
    query_res = snowflake_client.query_singlepoints(basis="sTO-3g")
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    query_res = snowflake_client.query_singlepoints(basis=[None])
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    query_res = snowflake_client.query_singlepoints(basis="")
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    # query for method
    query_res = snowflake_client.query_singlepoints(method=["b3lyP"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    # driver
    query_res = snowflake_client.query_singlepoints(driver=[SinglepointDriver.energy])
    query_res_l = list(query_res)
    assert len(query_res_l) == 3

    # keywords
    query_res = snowflake_client.query_singlepoints(keywords={})
    query_res_l = list(query_res)
    assert len(query_res_l) == 3

    query_res = snowflake_client.query_singlepoints(keywords={"maxiter": 100})
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    query_res = snowflake_client.query_singlepoints(keywords=[{"maxiter": 100}, {"something": 100}])
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    query_res = snowflake_client.query_singlepoints(keywords=[{}], program="rdkit")
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    query_res = snowflake_client.query_singlepoints(keywords={"maxiter": 100})
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    query_res = snowflake_client.query_singlepoints(keywords={"something": 100})
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    # Some empty queries
    query_res = snowflake_client.query_singlepoints(driver=[SinglepointDriver.properties])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    # Some empty queries
    query_res = snowflake_client.query_singlepoints(basis=["madeupbasis"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    # Query by default returns everything
    query_res = snowflake_client.query_singlepoints()
    query_res_l = list(query_res)
    assert len(query_res_l) == 3

    # Query by default (with a limit)
    query_res = snowflake_client.query_singlepoints(limit=1)
    query_res_l = list(query_res)
    assert len(query_res_l) == 1



"""add projects tables

Revision ID: a036e75160ea
Revises: db730926b1b9
Create Date: 2025-03-24 20:31:36.029509

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a036e75160ea"
down_revision = "db730926b1b9"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "project",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=100), nullable=False),
        sa.Column(
            "lname",
            sa.String(length=100),
            sa.Computed(
                "LOWER(name)",
            ),
            nullable=False,
        ),
        sa.Column("description", sa.String(), nullable=False),
        sa.Column("tagline", sa.String(), nullable=False),
        sa.Column("tags", sa.JSON(), nullable=False),
        sa.Column("default_compute_tag", sa.String(), nullable=False),
        sa.Column("default_compute_priority", sa.Integer(), nullable=False),
        sa.Column("extras", sa.JSON(), nullable=False),
        sa.Column("owner_user_id", sa.Integer(), nullable=True),
        sa.Column("owner_group_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["owner_group_id"],
            ["group.id"],
        ),
        sa.ForeignKeyConstraint(
            ["owner_user_id", "owner_group_id"],
            ["user_groups.user_id", "user_groups.group_id"],
        ),
        sa.ForeignKeyConstraint(
            ["owner_user_id"],
            ["user.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("lname", name="ux_project_project_type_lname"),
    )
    op.create_index("ix_project_owner_group_id", "project", ["owner_group_id"], unique=False)
    op.create_index("ix_project_owner_user_id", "project", ["owner_user_id"], unique=False)
    op.create_table(
        "project_attachment",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("project_id", sa.Integer(), nullable=False),
        sa.Column("attachment_type", sa.Enum("other", name="projectattachmenttype"), nullable=False),
        sa.Column("tags", sa.JSON(), nullable=False),
        sa.ForeignKeyConstraint(["id"], ["external_file.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(["project_id"], ["project.id"], ondelete="cascade"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index("ix_project_attachment_project_id", "project_attachment", ["project_id"], unique=False)
    op.create_table(
        "project_dataset",
        sa.Column("project_id", sa.Integer(), nullable=False),
        sa.Column("dataset_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dataset_id"],
            ["base_dataset.id"],
        ),
        sa.ForeignKeyConstraint(["project_id"], ["project.id"], ondelete="cascade"),
        sa.PrimaryKeyConstraint("project_id", "dataset_id"),
    )
    op.create_index("ix_project_dataset_dataset_id", "project_dataset", ["dataset_id"], unique=False)
    op.create_table(
        "project_internal_job",
        sa.Column("internal_job_id", sa.Integer(), nullable=False),
        sa.Column("project_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["internal_job_id"], ["internal_jobs.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(["project_id"], ["project.id"], ondelete="cascade"),
        sa.PrimaryKeyConstraint("internal_job_id", "project_id"),
    )
    op.create_table(
        "project_molecule",
        sa.Column("project_id", sa.Integer(), nullable=False),
        sa.Column("molecule_id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column(
            "lname",
            sa.String(length=100),
            sa.Computed(
                "LOWER(name)",
            ),
            nullable=False,
        ),
        sa.Column("description", sa.String(), nullable=False),
        sa.Column("tags", sa.JSON(), nullable=False),
        sa.ForeignKeyConstraint(
            ["molecule_id"],
            ["molecule.id"],
        ),
        sa.ForeignKeyConstraint(["project_id"], ["project.id"], ondelete="cascade"),
        sa.PrimaryKeyConstraint("project_id", "lname"),
    )
    op.create_index("ix_project_molecule_molecule_id", "project_molecule", ["molecule_id"], unique=False)
    op.create_table(
        "project_record",
        sa.Column("project_id", sa.Integer(), nullable=False),
        sa.Column("record_id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column(
            "lname",
            sa.String(length=100),
            sa.Computed(
                "LOWER(name)",
            ),
            nullable=False,
        ),
        sa.Column("description", sa.String(), nullable=False),
        sa.Column("tags", sa.JSON(), nullable=False),
        sa.ForeignKeyConstraint(["project_id"], ["project.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(
            ["record_id"],
            ["base_record.id"],
        ),
        sa.PrimaryKeyConstraint("project_id", "lname"),
    )
    op.create_index("ix_project_record_record_id", "project_record", ["record_id"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_project_record_record_id", table_name="project_record")
    op.drop_table("project_record")
    op.drop_index("ix_project_molecule_molecule_id", table_name="project_molecule")
    op.drop_table("project_molecule")
    op.drop_table("project_internal_job")
    op.drop_index("ix_project_dataset_dataset_id", table_name="project_dataset")
    op.drop_table("project_dataset")
    op.drop_index("ix_project_attachment_project_id", table_name="project_attachment")
    op.drop_table("project_attachment")
    op.drop_index("ix_project_owner_user_id", table_name="project")
    op.drop_index("ix_project_owner_group_id", table_name="project")
    op.drop_table("project")
    # ### end Alembic commands ###



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

try:
    from pydantic.v1 import ValidationError
except ImportError:
    from pydantic import ValidationError

from qcfractal.components.singlepoint.testing_helpers import submit_procedure_data
from qcportal import PortalRequestError
from qcportal.managers import ManagerName, ManagerStatusEnum
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake


def test_manager_mclient_activate(snowflake: QCATestingSnowflake):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")

    # UUID is different
    mname2 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5679")

    mclient1 = snowflake.manager_client(mname1)
    mclient2 = snowflake.manager_client(mname2)

    time_0 = now_at_utc()
    mclient1.activate(
        manager_version="v2.0",
        programs={"qcengine": ["unknown"], "program1": ["v3.0"]},
        compute_tags=["tag1", "tag2"],
    )

    time_1 = now_at_utc()
    mclient2.activate(
        manager_version="v2.0",
        programs={"qcengine": ["unknown"], "program1": ["v3.0"]},
        compute_tags=["tag1"],
    )
    time_2 = now_at_utc()

    name1 = mname1.fullname
    name2 = mname2.fullname

    client = snowflake.client()
    manager = client.get_managers([name2, name1])
    assert len(manager) == 2

    assert manager[0].name == name2
    assert manager[0].compute_tags == ["tag1"]
    assert manager[0].status == ManagerStatusEnum.active
    assert manager[0].created_on > time_1
    assert manager[0].modified_on > time_1
    assert manager[0].created_on < time_2
    assert manager[0].modified_on < time_2

    assert manager[1].name == name1
    assert manager[1].cluster == "test_cluster"
    assert manager[1].hostname == "a_host"
    assert manager[1].compute_tags == ["tag1", "tag2"]
    assert manager[1].status == ManagerStatusEnum.active
    assert manager[1].created_on > time_0
    assert manager[1].modified_on > time_0
    assert manager[1].created_on < time_1
    assert manager[1].modified_on < time_1


def test_manager_mclient_activate_normalize(snowflake: QCATestingSnowflake):
    # Activation where tags & programs are mixed case
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")

    mclient1 = snowflake.manager_client(mname1)
    client = snowflake.client()

    mclient1.activate(
        manager_version="v2.0",
        programs={"qcengine": ["unknown"], "program1": ["v3.0"], "PROgRam2": ["v4.0"], "PROGRAM4": ["v5.0-AB"]},
        compute_tags=["tag1", "taG3", "tAg2", "TAG3", "TAG1"],
    )

    manager = client.get_managers(mname1.fullname)
    assert manager.compute_tags == ["tag1", "tag3", "tag2"]
    assert manager.programs == {
        "qcengine": ["unknown"],
        "program1": ["v3.0"],
        "program2": ["v4.0"],
        "program4": ["v5.0-ab"],
    }


def test_manager_mclient_activate_notags(snowflake: QCATestingSnowflake):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")

    mclient1 = snowflake.manager_client(mname1)

    with pytest.raises(ValidationError, match=r"field contains no non-zero-length tags") as err:
        mclient1.activate(
            manager_version="v2.0",
            programs={"qcengine": ["unknown"], "program1": ["v3.0"]},
            compute_tags=[],
        )

    with pytest.raises(ValidationError, match=r"field contains no non-zero-length tags") as err:
        mclient1.activate(
            manager_version="v2.0",
            programs={"qcengine": ["unknown"], "program1": ["v3.0"]},
            compute_tags=[""],
        )


def test_manager_mclient_activate_noprogs(snowflake: QCATestingSnowflake):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")

    mclient1 = snowflake.manager_client(mname1)

    with pytest.raises(ValidationError, match=r"field contains no non-zero-length programs") as err:
        mclient1.activate(
            manager_version="v2.0",
            programs={},
            compute_tags=["tag1"],
        )

    with pytest.raises(ValidationError, match=r"field contains no non-zero-length programs") as err:
        mclient1.activate(
            manager_version="v2.0",
            programs={"": ["unknown"]},
            compute_tags=["tag1"],
        )


def test_manager_mclient_activate_duplicate(snowflake: QCATestingSnowflake):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")

    mclient1 = snowflake.manager_client(mname1)
    mclient1.activate(
        manager_version="v2.0",
        programs={"qcengine": ["unknown"], "program1": ["v3.0"]},
        compute_tags=["tag1"],
    )

    with pytest.raises(PortalRequestError, match=r"already exists") as err:
        mclient1.activate(
            manager_version="v2.0",
            programs={"qcengine": ["unknown"], "program1": ["v3.0"]},
            compute_tags=["tag1"],
        )


def test_manager_mclient_deactivate(snowflake: QCATestingSnowflake):
    client = snowflake.client()

    id1, _ = submit_procedure_data(snowflake.get_storage_socket(), "sp_psi4_benzene_energy_1", "tag1")

    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")
    mprog1 = {"qcengine": ["unknown"], "psi4": ["v3.0"]}

    # UUID is different
    mname2 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5679")
    mprog2 = {"qcengine": ["unknown"], "psi4": ["v3.0"]}

    mclient1 = snowflake.manager_client(mname1)
    mclient2 = snowflake.manager_client(mname2)

    mclient1.activate(
        manager_version="v2.0",
        programs=mprog1,
        compute_tags=["tag1", "tag2"],
    )

    mclient2.activate(
        manager_version="v2.0",
        programs=mprog2,
        compute_tags=["tag1"],
    )

    name1 = mname1.fullname
    name2 = mname2.fullname

    # client2 claims tasks
    mclient2.claim(mprog2, ["tag1"], 1)
    mclient2.deactivate(total_cpu_hours=1.0, active_tasks=1, active_cores=2, active_memory=7.0)

    manager = client.get_managers([name1, name2])
    assert manager[0].name == name1
    assert manager[0].status == ManagerStatusEnum.active
    assert manager[1].name == name2
    assert manager[1].status == ManagerStatusEnum.inactive
    assert manager[1].total_cpu_hours == 1.0
    assert manager[1].active_tasks == 1
    assert manager[1].active_cores == 2
    assert manager[1].active_memory == 7.0
    assert manager[1].claimed == 1

    record = client.get_records(id1)
    assert record.status == "waiting"
    assert record.manager_name is None


def test_manager_mclient_deactivate_deactivated(snowflake: QCATestingSnowflake):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")

    mclient1 = snowflake.manager_client(mname1)

    time_0 = now_at_utc()
    mclient1.activate(
        manager_version="v2.0",
        programs={"qcengine": ["unkonwn"], "program1": ["v3.0"]},
        compute_tags=["tag1", "tag2"],
    )

    mclient1.deactivate(total_cpu_hours=2.0, active_tasks=1, active_cores=2, active_memory=7.0)

    with pytest.raises(PortalRequestError, match=r"is not active") as err:
        mclient1.deactivate(total_cpu_hours=2.0, active_tasks=1, active_cores=2, active_memory=7.0)


def test_manager_mclient_heartbeat(snowflake: QCATestingSnowflake):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")

    mclient1 = snowflake.manager_client(mname1)

    time_0 = now_at_utc()
    mclient1.activate(
        manager_version="v2.0",
        programs={"qcengine": ["unknown"], "program1": ["v3.0"]},
        compute_tags=["tag1", "tag2"],
    )

    time_1 = now_at_utc()

    name1 = mname1.fullname

    client = snowflake.client()
    manager = client.get_managers(name1)

    # Now do a heartbeat
    mclient1.heartbeat(total_cpu_hours=5.678, active_tasks=3, active_cores=10, active_memory=3.45)

    time_2 = now_at_utc()

    manager = client.get_managers(name1)

    # Was the data stored in the manager
    assert manager.total_cpu_hours == 5.678
    assert manager.active_tasks == 3
    assert manager.active_cores == 10
    assert manager.active_memory == 3.45
    assert manager.modified_on > time_1
    assert manager.modified_on < time_2

    # Now do another heartbeat
    mclient1.heartbeat(
        total_cpu_hours=2 * 5.678,
        active_tasks=2 * 3,
        active_cores=2 * 10,
        active_memory=2 * 3.45,
    )

    time_3 = now_at_utc()

    manager = client.get_managers(name1)

    # Was the data stored in the manager
    assert manager.total_cpu_hours == 2 * 5.678
    assert manager.active_tasks == 2 * 3
    assert manager.active_cores == 2 * 10
    assert manager.active_memory == 2 * 3.45
    assert manager.modified_on > time_2
    assert manager.modified_on < time_3


def test_manager_mclient_heartbeat_deactivated(snowflake: QCATestingSnowflake):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")

    mclient1 = snowflake.manager_client(mname1)

    mclient1.activate(
        manager_version="v2.0",
        programs={"qcengine": ["unknown"], "program1": ["v3.0"]},
        compute_tags=["tag1", "tag2"],
    )

    mclient1.deactivate(total_cpu_hours=2.0, active_tasks=1, active_cores=2, active_memory=7.0)

    with pytest.raises(PortalRequestError, match=r"is not active") as err:
        mclient1.heartbeat(total_cpu_hours=5.678, active_tasks=3, active_cores=10, active_memory=3.45)



"""Add tag to sort task_queue sort index

Revision ID: f31c7897345f
Revises: 34d57d259c11
Create Date: 2024-04-09 09:50:13.771231

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f31c7897345f"
down_revision = "34d57d259c11"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_task_queue_sort", table_name="task_queue")
    op.create_index(
        "ix_task_queue_sort",
        "task_queue",
        [sa.text("priority DESC"), sa.text("sort_date ASC"), sa.text("id ASC"), "tag"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_task_queue_sort", table_name="task_queue")
    op.create_index("ix_task_queue_sort", "task_queue", [sa.text("priority DESC"), "sort_date", "id"], unique=False)
    # ### end Alembic commands ###



from __future__ import annotations

import copy
import logging
from typing import TYPE_CHECKING

from sqlalchemy import select, literal, insert

from qcfractal.components.torsiondrive.record_db_models import TorsiondriveRecordORM
from qcportal.metadata_models import InsertMetadata, InsertCountsMetadata
from qcportal.record_models import PriorityEnum
from qcportal.torsiondrive import TorsiondriveDatasetNewEntry, TorsiondriveSpecification
from .dataset_db_models import (
    TorsiondriveDatasetORM,
    TorsiondriveDatasetSpecificationORM,
    TorsiondriveDatasetEntryORM,
    TorsiondriveDatasetMoleculeORM,
    TorsiondriveDatasetRecordItemORM,
)
from ..base_dataset_socket import BaseDatasetSocket

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import Optional, Sequence, Iterable, Tuple


class TorsiondriveDatasetSocket(BaseDatasetSocket):
    # Used by the base class
    dataset_orm = TorsiondriveDatasetORM
    specification_orm = TorsiondriveDatasetSpecificationORM
    entry_orm = TorsiondriveDatasetEntryORM
    record_item_orm = TorsiondriveDatasetRecordItemORM
    record_orm = TorsiondriveRecordORM

    def __init__(self, root_socket: SQLAlchemySocket):
        BaseDatasetSocket.__init__(
            self,
            root_socket,
        )

        self._logger = logging.getLogger(__name__)

    def _add_specification(
        self, session: Session, specification: TorsiondriveSpecification
    ) -> Tuple[InsertMetadata, Optional[int]]:
        return self.root_socket.records.torsiondrive.add_specification(specification, session=session)

    def _create_entries(self, session: Session, dataset_id: int, new_entries: Sequence[TorsiondriveDatasetNewEntry]):
        all_entries = []
        for entry in new_entries:
            meta, mol_ids = self.root_socket.molecules.add_mixed(entry.initial_molecules, session=session)

            new_ent_mols = [TorsiondriveDatasetMoleculeORM(molecule_id=mid) for mid in mol_ids]

            new_ent = TorsiondriveDatasetEntryORM(
                dataset_id=dataset_id,
                name=entry.name,
                comment=entry.comment,
                additional_keywords=entry.additional_keywords,
                additional_optimization_keywords=entry.additional_optimization_keywords,
                attributes=entry.attributes,
                initial_molecules_assoc=new_ent_mols,
            )

            all_entries.append(new_ent)

        return all_entries

    def _submit(
        self,
        session: Session,
        dataset_id: int,
        entry_orm: Iterable[TorsiondriveDatasetEntryORM],
        spec_orm: Iterable[TorsiondriveDatasetSpecificationORM],
        existing_records: Iterable[Tuple[str, str]],
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user_id: Optional[int],
        find_existing: bool,
    ) -> InsertCountsMetadata:

        n_inserted = 0
        n_existing = 0

        for spec in spec_orm:
            td_spec_obj = spec.specification.to_model(TorsiondriveSpecification)
            td_spec_input_dict = td_spec_obj.dict()

            for entry in entry_orm:
                if (entry.name, spec.name) in existing_records:
                    continue

                new_td_spec = copy.deepcopy(td_spec_input_dict)
                new_td_spec["keywords"].update(entry.additional_keywords)
                new_td_spec["optimization_specification"]["keywords"].update(entry.additional_optimization_keywords)

                td_spec = TorsiondriveSpecification(
                    optimization_specification=new_td_spec["optimization_specification"],
                    keywords=new_td_spec["keywords"],
                )

                meta, td_ids = self.root_socket.records.torsiondrive.add(
                    initial_molecules=[entry.initial_molecule_ids],
                    td_spec=td_spec,
                    as_service=True,
                    compute_tag=compute_tag,
                    compute_priority=compute_priority,
                    creator_user=creator_user_id,
                    find_existing=find_existing,
                    session=session,
                )

                if meta.success:
                    rec = TorsiondriveDatasetRecordItemORM(
                        dataset_id=dataset_id,
                        entry_name=entry.name,
                        specification_name=spec.name,
                        record_id=td_ids[0],
                    )
                    session.add(rec)

                n_inserted += meta.n_inserted
                n_existing += meta.n_existing

        return InsertCountsMetadata(n_inserted=n_inserted, n_existing=n_existing)

    def _copy_entries(
        self,
        session: Session,
        source_dataset_id: int,
        destination_dataset_id: int,
        entry_names: Optional[Iterable[str]] = None,
    ):

        select_stmt = select(
            literal(destination_dataset_id),
            self.entry_orm.name,
            self.entry_orm.comment,
            self.entry_orm.additional_keywords,
            self.entry_orm.additional_optimization_keywords,
            self.entry_orm.attributes,
        )

        select_stmt = select_stmt.where(self.entry_orm.dataset_id == source_dataset_id)

        if entry_names is not None:
            select_stmt = select_stmt.where(self.entry_orm.name.in_(entry_names))

        stmt = insert(self.entry_orm)
        stmt = stmt.from_select(
            [
                self.entry_orm.dataset_id,
                self.entry_orm.name,
                self.entry_orm.comment,
                self.entry_orm.additional_keywords,
                self.entry_orm.additional_optimization_keywords,
                self.entry_orm.attributes,
            ],
            select_stmt,
        )

        session.execute(stmt)

        # Now do the molecules (stored in a separate table)
        select_stmt = select(
            literal(destination_dataset_id),
            TorsiondriveDatasetMoleculeORM.entry_name,
            TorsiondriveDatasetMoleculeORM.molecule_id,
        )

        select_stmt = select_stmt.where(TorsiondriveDatasetMoleculeORM.dataset_id == source_dataset_id)

        if entry_names is not None:
            select_stmt = select_stmt.where(TorsiondriveDatasetMoleculeORM.entry_name.in_(entry_names))

        stmt = insert(TorsiondriveDatasetMoleculeORM)
        stmt = stmt.from_select(
            [
                TorsiondriveDatasetMoleculeORM.dataset_id,
                TorsiondriveDatasetMoleculeORM.entry_name,
                TorsiondriveDatasetMoleculeORM.molecule_id,
            ],
            select_stmt,
        )

        session.execute(stmt)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from qcelemental.molutil import order_molecular_formula
from sqlalchemy.orm import load_only
from sqlalchemy.orm.attributes import flag_modified
from sqlalchemy.sql import select, and_, or_

from qcfractal.db_socket.helpers import (
    insert_general,
    delete_general,
    insert_mixed_general,
    get_general,
)
from qcportal.exceptions import MissingDataError
from qcportal.metadata_models import (
    InsertMetadata,
    DeleteMetadata,
    UpdateMetadata,
)
from qcportal.molecules import Molecule, MoleculeIdentifiers, MoleculeQueryFilters, MoleculeUploadOptions
from .db_models import MoleculeORM
from .from_files import molecules_from_files

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import List, Union, Tuple, Optional, Sequence, Dict, Any


class MoleculeSocket:
    """
    Socket for managing/querying molecules
    """

    def __init__(self, root_socket: SQLAlchemySocket):
        self.root_socket = root_socket
        self._logger = logging.getLogger(__name__)

    def add(
        self, molecules: Sequence[Molecule], *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, List[int]]:
        """
        Add molecules to the database

        This checks if the molecule already exists in the database via its hash. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        Parameters
        ----------
        molecules
            Molecule data to add to the session
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of Molecule ids.
            The ids will be in the order of the input molecules.
        """

        ###############################################################################
        # Exceptions in this function would usually be a programmer error, as any
        # valid Molecule object should be insertable into the database
        ###############################################################################

        molecule_orm = [MoleculeORM.from_model(x) for x in molecules]

        with self.root_socket.optional_session(session) as session:
            meta, added_ids = insert_general(
                session,
                molecule_orm,
                (MoleculeORM.molecule_hash,),
                (MoleculeORM.id,),
                use_unique=True,
            )

        # added_ids is a list of tuple, with each tuple only having one value. Flatten that out
        return meta, [x[0] for x in added_ids]

    def add_from_files(
        self, file_list: List[Tuple[str, str]], options: MoleculeUploadOptions, *, session: Optional[Session] = None
    ) -> Tuple[Dict[str, List[Tuple[str, int]]], List[str]]:
        minfo, errors = molecules_from_files(file_list, options)

        ret: Dict[str, List[Tuple[str, int]]] = {}
        for k, v in minfo.items():
            meta, molecule_ids = self.add([x[1] for x in v], session=session)
            if not meta.success:
                ret[k] = []  # Just so people can see nothing was inserted
                errors.append(meta.error_string)
            else:
                ret[k] = [(fname, mid) for (fname, _), mid in zip(v, molecule_ids)]

        return ret, errors

    def get(
        self,
        molecule_id: Sequence[int],
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
        missing_ok: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> List[Optional[Dict[str, Any]]]:
        """
        Obtain molecules with specified IDs from the database

        Parameters
        ----------
        molecule_id
            A list or other sequence of molecule IDs
        include
            Which fields of the molecule to return. Default is to return all fields.
        exclude
            Remove these fields from the return. Default is to return all fields.
        missing_ok
            If set to True, then missing molecules will be tolerated, and the returned list of
            Molecules will contain None for the corresponding IDs that were not found.
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            List of molecule data (as dictionaries) in the same order as the given ids.
            If missing_ok is True, then this list will contain None where the molecule was missing.
        """

        with self.root_socket.optional_session(session, True) as session:
            return get_general(session, MoleculeORM, MoleculeORM.id, molecule_id, include, exclude, missing_ok)

    def add_mixed(
        self, molecule_data: Sequence[Union[int, Molecule]], *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Add mixed molecules and ids to the database.

        This function can take both Molecule objects and molecule ids. If a molecule id is given
        in the list, then it is checked to make sure it exists. If it does not exist, then it will be
        marked as an error in the returned metadata and the corresponding entry in the returned
        list of IDs will be None.

        If a Molecule object is given, it will be added to the database if it does not already exist
        in the database (based on the hash) and the existing ID will be returned. Otherwise, the new
        ID will be returned.

        Parameters
        ----------
        molecule_data
            Molecule data to add. Can be a mix of IDs and Molecule objects
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of Molecule ids.
            The ids will be in the order of the input molecules.
        """

        molecule_orm: List[Union[int, MoleculeORM]] = [
            x if isinstance(x, int) else MoleculeORM.from_model(x) for x in molecule_data
        ]

        with self.root_socket.optional_session(session) as session:
            meta, all_ids = insert_mixed_general(
                session,
                MoleculeORM,
                molecule_orm,
                MoleculeORM.id,
                (MoleculeORM.molecule_hash,),
                (MoleculeORM.id,),
                use_unique=True,
            )

        # added_ids is a list of tuple, with each tuple only having one value. Flatten that out
        return meta, [x[0] if x is not None else None for x in all_ids]

    def delete(self, molecule_id: Sequence[int], *, session: Optional[Session] = None) -> DeleteMetadata:
        """
        Removes molecules with the given ids from the database

        Parameters
        ----------
        molecule_id
            IDs of the molecules to remove
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about what was deleted and any errors that occurred
        """

        id_lst = [(x,) for x in molecule_id]

        with self.root_socket.optional_session(session) as session:
            return delete_general(session, MoleculeORM, MoleculeORM.id, id_lst)

    def query(
        self,
        query_data: MoleculeQueryFilters,
        *,
        session: Optional[Session] = None,
    ) -> List[int]:
        """
        General query of molecules in the database

        All search criteria are merged via 'and'. Therefore, records will only
        be found that match all the criteria.

        Parameters
        ----------
        query_data
            Fields/filters to query for
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            A list of molecule ids that were found in the database.
        """

        molecular_formula = query_data.molecular_formula
        identifiers = query_data.identifiers

        if molecular_formula is not None:
            try:
                # Make sure the molecular formulae are in the proper element order
                molecular_formula = [order_molecular_formula(form) for form in query_data.molecular_formula]
            except ValueError:
                # Probably, the user provided an invalid chemical formula
                pass

        and_query = []
        if query_data.molecule_id is not None:
            and_query.append(MoleculeORM.id.in_(query_data.molecule_id))
        if query_data.molecule_hash is not None:
            and_query.append(MoleculeORM.molecule_hash.in_(query_data.molecule_hash))
        if molecular_formula is not None:
            # Add it to the identifiers query
            if identifiers is None:
                identifiers = {"molecular_formula": list(molecular_formula)}
            else:
                identifiers["molecular_formula"] = list(molecular_formula)
        if identifiers is not None:
            for i_name, i_values in identifiers.items():
                or_query = []
                for v in i_values:
                    or_query.append(MoleculeORM.identifiers.contains({i_name: v}))
                and_query.append(or_(False, *or_query))

        with self.root_socket.optional_session(session, True) as session:
            stmt = select(MoleculeORM.id).where(and_(True, *and_query))

            if query_data.cursor is not None:
                stmt = stmt.where(MoleculeORM.id < query_data.cursor)

            stmt = stmt.order_by(MoleculeORM.id.desc())
            stmt = stmt.limit(query_data.limit)
            stmt = stmt.distinct(MoleculeORM.id)
            molecule_ids = session.execute(stmt).scalars().all()

        return molecule_ids

    def modify(
        self,
        molecule_id: int,
        name: Optional[str] = None,
        comment: Optional[str] = None,
        identifiers: Optional[MoleculeIdentifiers] = None,
        overwrite_identifiers: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> UpdateMetadata:
        """
        Modify molecule information in the database

        This is only capable of updating the name, comment, and identifiers fields (except molecule_hash
        and molecular formula).

        If a molecule with that id does not exist, an exception is raised

        Parameters
        ----------
        molecule_id
            Molecule ID of the molecule to modify
        name
            New name for the molecule. If None, name is not changed.
        comment
            New comment for the molecule. If None, comment is not changed
        identifiers
            New identifiers for the molecule
        overwrite_identifiers
            If True, the identifiers of the molecule are set to be those given exactly (ie, identifiers
            that exist in the DB but not in the new set will be removed). Otherwise, the new set of
            identifiers is merged into the existing ones. Note that molecule_hash and molecular_formula
            are never removed.
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the modification/update.
        """

        with self.root_socket.optional_session(session) as session:
            stmt = select(MoleculeORM).where(or_(MoleculeORM.id == molecule_id)).with_for_update()
            stmt = stmt.options(
                load_only(MoleculeORM.id, MoleculeORM.name, MoleculeORM.comment, MoleculeORM.identifiers)
            )
            mol = session.execute(stmt).scalar_one_or_none()

            if mol is None:
                raise MissingDataError(f"Molecule with id {molecule_id} not found in the database")

            if name is not None:
                mol.name = name
            if comment is not None:
                mol.comment = comment
            if identifiers is not None:
                update_dict = {
                    "molecule_hash": mol.identifiers["molecule_hash"],
                    "molecular_formula": mol.identifiers["molecular_formula"],
                }

                # Changing identifiers is a bit sensitive, so validate again
                identifiers = identifiers.copy(update=update_dict)

                if overwrite_identifiers:
                    # Always keep hash & formula
                    mol.identifiers = identifiers.dict(exclude_unset=True, exclude_defaults=True, exclude_none=True)
                else:
                    id_dict = identifiers.dict(exclude_unset=True, exclude_defaults=True)
                    mol.identifiers.update(id_dict)

                    # sqlalchemy cannot track changes in json
                    flag_modified(mol, "identifiers")

        return UpdateMetadata(updated_idx=[0])



from __future__ import annotations

from typing import TYPE_CHECKING, Tuple, Optional, Dict, List, Union, Any

try:
    import pydantic.v1 as pydantic
except ImportError:
    import pydantic
from qcarchivetesting.helpers import read_procedure_data, read_record_data, find_test_data
from qcelemental.models import (
    Molecule,
    FailedOperation,
    ComputeError,
    AtomicResult as QCEl_AtomicResult,
    OptimizationResult as QCEl_OptimizationResult,
)

from qcfractal.components.reaction.record_db_models import ReactionRecordORM
from qcfractal.testing_helpers import run_service
from qcportal.reaction import ReactionSpecification, ReactionKeywords, ReactionRecord
from qcportal.record_models import PriorityEnum, RecordStatusEnum, RecordTask
from qcportal.singlepoint import SinglepointProtocols, QCSpecification

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from qcportal.managers import ManagerName

all_test_data = find_test_data("rxn_*")
all_includes = ["components", "molecule", "comments", "initial_molecule", "final_molecule"]

test_specs = [
    ReactionSpecification(
        program="reaction",
        singlepoint_specification=QCSpecification(
            program="prog1",
            driver="energy",
            method="b3lyp",
            basis="6-31G*",
            keywords={"k": "value"},
            protocols=SinglepointProtocols(wavefunction="all"),
        ),
        keywords=ReactionKeywords(),
    ),
    ReactionSpecification(
        program="reaction",
        singlepoint_specification=QCSpecification(
            program="Prog2", driver="energy", method="Hf", basis="def2-TZVP", keywords={"k": "v"}
        ),
        keywords=ReactionKeywords(),
    ),
]


def compare_reaction_specs(
    input_spec: Union[ReactionSpecification, Dict[str, Any]],
    output_spec: Union[ReactionSpecification, Dict[str, Any]],
) -> bool:
    if isinstance(input_spec, dict):
        input_spec = ReactionSpecification(**input_spec)
    if isinstance(output_spec, dict):
        output_spec = ReactionSpecification(**output_spec)

    return input_spec == output_spec


def generate_task_key(task: RecordTask):
    inp_data = task.function_kwargs["input_data"]

    if inp_data["schema_name"] == "qcschema_optimization_input":
        record_type = "optimization"
        mol_hash = inp_data["initial_molecule"]["identifiers"]["molecule_hash"]
    elif inp_data["schema_name"] == "qcschema_input":
        record_type = "singlepoint"
        mol_hash = inp_data["molecule"]["identifiers"]["molecule_hash"]
    else:
        raise RuntimeError(f"Unknown result type: {inp_data['schema_name']}")

    return record_type + "|" + mol_hash


def load_procedure_data(
    name: str,
) -> Tuple[
    ReactionSpecification, List[Tuple[float, Molecule]], Dict[str, Union[QCEl_AtomicResult, QCEl_OptimizationResult]]
]:
    data = read_procedure_data(name)

    return (
        pydantic.parse_obj_as(ReactionSpecification, data["specification"]),
        pydantic.parse_obj_as(List[Tuple[float, Molecule]], data["stoichiometry"]),
        pydantic.parse_obj_as(Dict[str, Union[QCEl_AtomicResult, QCEl_OptimizationResult]], data["results"]),
    )


def load_record_data(name: str) -> ReactionRecord:
    data = read_record_data(name)
    return ReactionRecord(**data)


def submit_procedure_data(
    storage_socket: SQLAlchemySocket,
    name: str,
    compute_tag: Optional[str] = "*",
    compute_priority: PriorityEnum = PriorityEnum.normal,
) -> Tuple[int, Dict[str, Union[QCEl_AtomicResult, QCEl_OptimizationResult]]]:
    input_spec, stoich, result = load_procedure_data(name)
    meta, record_ids = storage_socket.records.reaction.add(
        [stoich], input_spec, compute_tag, compute_priority, None, True
    )
    assert meta.success
    assert len(record_ids) == 1
    assert meta.n_inserted == 1

    return record_ids[0], result


def run_procedure_data(
    storage_socket: SQLAlchemySocket,
    manager_name: ManagerName,
    name: str,
    compute_tag: Optional[str] = "*",
    compute_priority: PriorityEnum = PriorityEnum.normal,
    end_status: RecordStatusEnum = RecordStatusEnum.complete,
):
    record_id, result = submit_procedure_data(storage_socket, name, compute_tag, compute_priority)

    with storage_socket.session_scope() as session:
        record = session.get(ReactionRecordORM, record_id)
        assert record.status == RecordStatusEnum.waiting

    if end_status == RecordStatusEnum.error:
        failed_op = FailedOperation(
            error=ComputeError(error_type="test_error", error_message="this is just a test error"),
        )
        result = {x: failed_op for x in result}

    finished, n_optimizations = run_service(storage_socket, manager_name, record_id, generate_task_key, result, 200)
    assert finished

    with storage_socket.session_scope() as session:
        record = session.get(ReactionRecordORM, record_id)
        assert record.status == end_status

    return record_id



from __future__ import annotations

from typing import Optional, Iterable, Dict, Any

from sqlalchemy import Column, Integer, TIMESTAMP, String, Float, BigInteger, JSON, Index, CHAR, ForeignKey
from sqlalchemy.dialects.postgresql import INET
from sqlalchemy.orm import relationship

from qcfractal.components.auth.db_models import UserORM, UserIDMapSubquery
from qcfractal.db_socket import BaseORM
from qcportal.utils import now_at_utc


class ServerStatsMetadataORM(BaseORM):
    __tablename__ = "server_stats_metadata"

    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    date_value = Column(TIMESTAMP(timezone=True), nullable=False)


class AccessLogORM(BaseORM):
    """
    Table for storing a log of accesses/requests
    """

    __tablename__ = "access_log"

    id = Column(Integer, primary_key=True)
    timestamp = Column(TIMESTAMP(timezone=True), nullable=False, default=now_at_utc)
    method = Column(String, nullable=False)
    module = Column(String, nullable=True)
    full_uri = Column(String, nullable=True)

    request_duration = Column(Float, nullable=False)
    request_bytes = Column(BigInteger, nullable=False)
    response_bytes = Column(BigInteger, nullable=False)

    user_id = Column(Integer, ForeignKey(UserORM.id), nullable=True)

    user = relationship(
        UserIDMapSubquery,
        foreign_keys=[user_id],
        primaryjoin="AccessLogORM.user_id == UserIDMapSubquery.id",
        lazy="selectin",
    )

    # user info
    ip_address = Column(INET)
    user_agent = Column(String)

    # extra computed geo data
    country_code = Column(CHAR(2))
    subdivision = Column(String)
    city = Column(String)
    ip_lat = Column(Float)
    ip_long = Column(Float)

    __table_args__ = (
        Index("ix_access_log_timestamp", "timestamp", postgresql_using="brin"),
        Index("ix_access_log_module", "module"),
        Index("ix_access_log_user_id", "user_id"),
    )

    _qcportal_model_excludes = ["user_id"]

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        d = BaseORM.model_dict(self, exclude)
        d["user"] = self.user.username if self.user is not None else None
        return d


class InternalErrorLogORM(BaseORM):
    """
    Table for storing internal errors

    Internal errors are usually not reported to the user, and are stored here instead
    for retrieval by an admin
    """

    __tablename__ = "internal_error_log"

    id = Column(Integer, primary_key=True)
    error_date = Column(TIMESTAMP(timezone=True), nullable=False, default=now_at_utc)
    qcfractal_version = Column(String, nullable=False)
    error_text = Column(String)
    user_id = Column(Integer, ForeignKey(UserORM.id), nullable=True)

    user = relationship(
        UserIDMapSubquery,
        foreign_keys=[user_id],
        primaryjoin="InternalErrorLogORM.user_id == UserIDMapSubquery.id",
        lazy="selectin",
    )

    request_path = Column(String)
    request_headers = Column(String)
    request_body = Column(String)

    __table_args__ = (
        Index("ix_internal_error_log_error_date", "error_date", postgresql_using="brin"),
        Index("ix_internal_error_log_user_id", "user_id"),
    )

    _qcportal_model_excludes = ["user_id"]

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        d = BaseORM.model_dict(self, exclude)
        d["user"] = self.user.username if self.user is not None else None
        return d


class MessageOfTheDayORM(BaseORM):
    """
    Table for storing the Message-of-the-Day
    """

    __tablename__ = "motd"

    id = Column(Integer, primary_key=True)
    motd = Column(String, nullable=False)



"""merge manybody branch with changes

Revision ID: d5988aa750ae
Revises: 3690c677f8d1, fd95035b773b
Create Date: 2025-01-14 10:49:27.547435

"""

# revision identifiers, used by Alembic.
revision = "d5988aa750ae"
down_revision = ("3690c677f8d1", "fd95035b773b")
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



"""Merge projects and manager migrations

Revision ID: 0c0b1dea52dc
Revises: a6a708376a1e, 93fa1a1b70f8
Create Date: 2025-04-27 12:47:56.091604

"""

# revision identifiers, used by Alembic.
revision = "0c0b1dea52dc"
down_revision = ("a6a708376a1e", "93fa1a1b70f8")
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



from __future__ import annotations

import importlib
import json
import logging
from typing import TYPE_CHECKING

import numpy as np
import sqlalchemy.orm.attributes
import tabulate

try:
    from pydantic.v1 import BaseModel, Extra
except ImportError:
    from pydantic import BaseModel, Extra

from sqlalchemy import select, func
from sqlalchemy.dialects.postgresql import array_agg, aggregate_order_by, DOUBLE_PRECISION, TEXT
from sqlalchemy.orm import lazyload, joinedload, selectinload, defer, undefer

from qcfractal.components.molecules.db_models import MoleculeORM
from qcfractal.components.services.db_models import ServiceQueueORM, ServiceDependencyORM
from qcfractal.components.singlepoint.record_db_models import QCSpecificationORM, SinglepointRecordORM
from qcfractal.db_socket.helpers import insert_general
from qcportal.exceptions import MissingDataError
from qcportal.metadata_models import InsertMetadata, InsertCountsMetadata
from qcportal.molecules import Molecule
from qcportal.neb import (
    NEBSpecification,
    NEBQueryFilters,
    NEBInput,
    NEBMultiInput,
    NEBRecord,
)
from qcportal.optimization import OptimizationSpecification
from qcportal.record_models import PriorityEnum, RecordStatusEnum, OutputTypeEnum
from qcportal.serialization import convert_numpy_recursive
from qcportal.singlepoint import QCSpecification
from qcportal.utils import capture_all_output, hash_dict, is_included
from .record_db_models import (
    NEBOptimizationsORM,
    NEBSpecificationORM,
    NEBSinglepointsORM,
    NEBInitialchainORM,
    NEBRecordORM,
)
from ..base_record_socket import BaseRecordSocket
from ..record_utils import append_output

# geometric package is optional
_geo_spec = importlib.util.find_spec("geometric")

if _geo_spec is not None:
    geometric = importlib.util.module_from_spec(_geo_spec)
    _geo_spec.loader.exec_module(geometric)

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import List, Dict, Tuple, Optional, Sequence, Any, Union

# Meaningless, but unique to neb
neb_insert_lock_id = 14600


class NEBServiceState(BaseModel):
    """
    This represents the current state of a NEB service
    """

    class Config:
        extra = Extra.forbid
        allow_mutation = True
        validate_assignment = True

    # These are stored as JSON (ie, dict encoded into a string)
    # This makes for faster loads and makes them somewhat tamper-proof

    nebinfo: dict
    keywords: dict
    optimized: bool
    tsoptimize: bool
    converged: bool
    iteration: int
    molecule_template: str
    tshessian: list


class NEBRecordSocket(BaseRecordSocket):
    # Used by the base class
    record_orm = NEBRecordORM
    record_input_type = NEBInput
    record_multi_input_type = NEBMultiInput

    def __init__(self, root_socket: SQLAlchemySocket):
        BaseRecordSocket.__init__(self, root_socket)
        self._logger = logging.getLogger(__name__)

    def available(self) -> bool:
        return _geo_spec is not None

    def initialize_service(
        self,
        session: Session,
        service_orm: ServiceQueueORM,
    ):
        neb_orm: NEBRecordORM = service_orm.record
        output = "\n\nCreated NEB calculation\n"
        spec: NEBSpecification = neb_orm.specification.to_model(NEBSpecification)
        keywords = spec.keywords.dict()
        table_rows = sorted(keywords.items())
        output += tabulate.tabulate(table_rows, headers=["NEB keywords", "value"])
        output += "\n\n"
        table_rows = sorted(spec.singlepoint_specification.dict().items())
        output += tabulate.tabulate(table_rows, headers=["NEB QC keywords", "value"])
        output += "\n\n"
        if bool(spec.optimization_specification):
            table_rows = sorted(spec.optimization_specification.qc_specification.dict().items())
            output += tabulate.tabulate(table_rows, headers=["OPT QC keywords", "value"])
        output += "\n\n"

        initial_chain: List[Dict[str, Any]] = [x.molecule.model_dict() for x in neb_orm.initial_chain]
        output += f"{keywords.get('images', 11)} images will be used to guess a transition state structure.\n"
        output += f"Molecular formula = {Molecule(**initial_chain[0]).get_molecular_formula()}\n"
        molecule_template = Molecule(**initial_chain[0]).dict(encoding="json")

        molecule_template.pop("geometry", None)
        molecule_template.pop("identifiers", None)
        molecule_template.pop("id", None)

        append_output(session, neb_orm, OutputTypeEnum.stdout, output)

        molecule_template_str = json.dumps(molecule_template)
        service_state = NEBServiceState(
            nebinfo={
                "elems": molecule_template.get("symbols"),
                "charge": molecule_template.get("molecular_charge", 0),
                "mult": molecule_template.get("molecular_multiplicity", 1),
            },
            iteration=0,
            keywords=keywords,
            optimized=keywords.get("optimize_endpoints"),
            tsoptimize=keywords.get("optimize_ts"),
            converged=False,
            tshessian=[],
            molecule_template=molecule_template_str,
        )

        service_orm.service_state = service_state.dict()
        sqlalchemy.orm.attributes.flag_modified(service_orm, "service_state")

    def iterate_service(
        self,
        session: Session,
        service_orm: ServiceQueueORM,
    ) -> bool:
        finished = False

        neb_orm: NEBRecordORM = service_orm.record
        # Always update with the current provenance
        neb_orm.compute_history[-1].provenance = {
            "creator": "neb",
            "version": geometric.__version__,
            "routine": "qcfractal.services.neb",
        }
        # Load the state from the service_state column
        spec: NEBSpecification = neb_orm.specification.to_model(NEBSpecification)
        params = spec.keywords.dict()
        service_state = NEBServiceState(**service_orm.service_state)
        molecule_template = json.loads(service_state.molecule_template)

        params["iteration"] = service_state.iteration
        output = ""
        if service_state.iteration == 0:
            service_state.converged = False
            initial_molecules = [x.molecule.to_model(Molecule) for x in neb_orm.initial_chain]

            if service_state.optimized:
                # First iteration, but we have been told to optimize the endpoints
                output += "\nFirst, optimizing the end points"
                self.submit_optimizations(session, service_orm, [initial_molecules[0], initial_molecules[-1]])
                service_state.optimized = False
            else:
                # Either no endpoint optimizations, or they have been done
                complete_opts = sorted(service_orm.dependencies, key=lambda x: x.extras["position"])

                # If there are completed opts, replace the endpoints with the optimized molecules
                if len(complete_opts) != 0:
                    if len(complete_opts) != 2:
                        raise RuntimeError(f"Expected 2 completed optimizations, got {len(complete_opts)}")

                    initial_molecules[0] = complete_opts[0].record.final_molecule.to_model(Molecule)
                    initial_molecules[-1] = complete_opts[-1].record.final_molecule.to_model(Molecule)

                with capture_all_output("geometric.nifty") as (rdout, _):
                    respaced_chain = geometric.qcf_neb.arrange(initial_molecules, params.get("align"))

                output += "\n" + rdout.getvalue()

                # Submit the first batch of singlepoint calculations
                self.submit_singlepoints(session, service_state, service_orm, respaced_chain)
                service_state.iteration += 1

        else:
            if not service_state.converged:
                # Returned task a nextchain computation
                if service_orm.dependencies and service_orm.dependencies[0].record.record_type == "servicesubtask":
                    newcoords, prev = service_orm.dependencies[0].record.results
                    service_state.nebinfo = prev

                    # Append the output
                    stdout = self.root_socket.records.service_subtask.get_single_output_uncompressed(
                        service_orm.dependencies[0].record.id,
                        service_orm.dependencies[0].record.compute_history[-1].id,
                        OutputTypeEnum.stdout,
                    )

                    output += stdout

                    # Delete the subtask - no longer needed
                    to_delete = service_orm.dependencies.pop()
                    self.root_socket.records.delete([to_delete.record_id], soft_delete=False, session=session)

                    # If nextchain returns None, then we are now converged
                    if newcoords is None:
                        service_state.converged = True
                    else:
                        next_chain = [Molecule(**molecule_template, geometry=geometry) for geometry in newcoords]
                        self.submit_singlepoints(session, service_state, service_orm, next_chain)
                        service_state.iteration += 1

                else:
                    complete_tasks = sorted(service_orm.dependencies, key=lambda x: x.extras["position"])
                    geometries = []
                    energies = []
                    gradients = []
                    for task in complete_tasks:
                        sp_record = task.record
                        mol_data = self.root_socket.molecules.get(
                            molecule_id=[sp_record.molecule_id], include=["geometry"], session=session
                        )
                        geometries.append(mol_data[0]["geometry"])
                        energies.append(sp_record.properties["return_energy"])
                        gradients.append(convert_numpy_recursive(sp_record.properties["return_result"], flatten=True))
                    service_state.nebinfo["geometry"] = convert_numpy_recursive(geometries, flatten=False)
                    service_state.nebinfo["energies"] = energies
                    service_state.nebinfo["gradients"] = gradients
                    service_state.nebinfo["params"] = params

                    with capture_all_output("geometric.nifty") as (rdout, _):
                        if service_state.iteration == 1:
                            newcoords, prev = geometric.qcf_neb.prepare(service_state.nebinfo)
                            service_state.nebinfo = prev

                            next_chain = [Molecule(**molecule_template, geometry=geometry) for geometry in newcoords]
                            self.submit_singlepoints(session, service_state, service_orm, next_chain)
                            service_state.iteration += 1
                        else:
                            self.submit_nextchain_subtask(session, service_state, service_orm)

                    output += "\n" + rdout.getvalue()

            # We are converged, but need to handle TS optimization
            if service_state.converged and service_state.tsoptimize:
                stmt = (
                    select(MoleculeORM)
                    .join(SinglepointRecordORM)
                    .join(NEBSinglepointsORM)
                    .where(NEBSinglepointsORM.neb_id == neb_orm.id)
                    .order_by(
                        NEBSinglepointsORM.chain_iteration.desc(),
                        SinglepointRecordORM.properties["return_energy"].cast(TEXT).cast(DOUBLE_PRECISION).desc(),
                    )
                    .limit(1)
                )

                TS_mol = session.execute(stmt).scalar_one_or_none()
                if TS_mol is None:
                    raise MissingDataError("MoleculeORM of a guessed transition state from NEB can't be found.")

                # Has the TS hessian calculation been completed?
                if len(service_state.tshessian) == 0:
                    output += (
                        "\nOptimizing the guessed transition state structure to locate a first-order saddle point.\n"
                        "Hessian will be calculated and passed to geomeTRIC."
                    )

                    self.submit_singlepoints(session, service_state, service_orm, [Molecule(**TS_mol.model_dict())])

                    # Mark that we are expecting the tshessian the next iteration
                    service_state.tshessian = [1]
                else:
                    # Hessian completed, do optimization
                    complete_sp = service_orm.dependencies[0]
                    service_orm.service_state["tshessian"] = complete_sp.record.properties["return_result"]
                    self.submit_optimizations(session, service_orm, [Molecule(**TS_mol.model_dict())])
                    service_state.tsoptimize = False

            elif service_state.converged:
                # Converged, but not waiting for anything else
                finished = True
                output += "\nNEB calculation is completed with %i iterations" % service_state.iteration

        append_output(session, neb_orm, OutputTypeEnum.stdout, output)
        service_orm.service_state = service_state.dict()
        sqlalchemy.orm.attributes.flag_modified(service_orm, "service_state")
        return finished

    def submit_optimizations(
        self,
        session: Session,
        service_orm: ServiceQueueORM,
        molecules: List[Molecule],
    ):
        neb_orm: NEBRecordORM = service_orm.record
        # delete all existing entries in the dependency list
        service_orm.dependencies = []
        service_state = NEBServiceState(**service_orm.service_state)
        neb_spec = neb_orm.specification.to_model(NEBSpecification)
        has_optimization = bool(neb_spec.optimization_specification)
        qc_spec = neb_orm.specification.singlepoint_specification.model_dict()
        if service_state.tsoptimize and service_state.converged:
            if has_optimization:
                opt_spec = neb_orm.specification.optimization_specification.model_dict()
                opt_spec["keywords"]["hessian"] = service_state.tshessian
                opt_spec["keywords"]["transition"] = True
                opt_spec["program"] = "geometric"
                opt_spec = OptimizationSpecification(
                    **opt_spec
                )  # neb_orm.specification.optimization_specification.to_model(OptimizationSpecification)

            else:
                opt_spec = OptimizationSpecification(
                    program="geometric",
                    qc_specification=QCSpecification(**qc_spec),
                    keywords={
                        "transition": True,
                        "coordsys": "tric",
                        "hessian": service_state.tshessian,
                    },
                )
            ts = True
        else:
            opt_spec = OptimizationSpecification(
                program="geometric",
                qc_specification=QCSpecification(**qc_spec),
                keywords={"coordsys": "tric"},
            )
            ts = False

        meta, opt_ids = self.root_socket.records.optimization.add(
            molecules,
            opt_spec,
            service_orm.compute_tag,
            service_orm.compute_priority,
            neb_orm.creator_user_id,
            service_orm.find_existing,
            session=session,
        )
        for pos, opt_id in enumerate(opt_ids):
            svc_dep = ServiceDependencyORM(record_id=opt_id, extras={"position": pos})
            opt_history = NEBOptimizationsORM(
                neb_id=service_orm.record_id,
                optimization_id=opt_id,
                position=pos,
                ts=ts,
            )
            service_orm.dependencies.append(svc_dep)
            neb_orm.optimizations.append(opt_history)

    def submit_singlepoints(
        self,
        session: Session,
        service_state: NEBServiceState,
        service_orm: ServiceQueueORM,
        chain: List[Molecule],
    ):
        neb_orm: NEBRecordORM = service_orm.record
        # delete all existing entries in the dependency list
        service_orm.dependencies = []
        neb_spec = neb_orm.specification.to_model(NEBSpecification)
        has_optimization = bool(neb_spec.optimization_specification)

        # Create a singlepoint input based on the multiple geometries
        qc_spec = neb_orm.specification.singlepoint_specification.model_dict()
        qc_spec["driver"] = "gradient"

        if service_state.converged and service_state.tsoptimize and len(service_state.tshessian) == 0:
            if has_optimization:
                opt_spec = neb_orm.specification.optimization_specification.model_dict()
                qc_spec = opt_spec["qc_specification"]
            qc_spec["driver"] = "hessian"

        meta, sp_ids = self.root_socket.records.singlepoint.add(
            chain,
            QCSpecification(**qc_spec),
            service_orm.compute_tag,
            service_orm.compute_priority,
            neb_orm.creator_user_id,
            service_orm.find_existing,
            session=session,
        )

        if not meta.success:
            raise RuntimeError("Error adding singlepoints - likely a developer error: " + meta.error_string)

        for pos, sp_id in enumerate(sp_ids):
            svc_dep = ServiceDependencyORM(
                record_id=sp_id,
                extras={"position": pos},
            )
            sp_history = NEBSinglepointsORM(
                neb_id=service_orm.record_id,
                singlepoint_id=sp_id,
                chain_iteration=service_state.iteration,
                position=pos,
            )

            service_orm.dependencies.append(svc_dep)
            neb_orm.singlepoints.append(sp_history)

    def submit_nextchain_subtask(
        self,
        session: Session,
        service_state: NEBServiceState,
        service_orm: ServiceQueueORM,
    ):
        neb_orm: NEBRecordORM = service_orm.record

        # delete all existing entries in the dependency list
        service_orm.dependencies = []

        meta, ids = self.root_socket.records.service_subtask.add(
            {"geometric": None},
            "geometric.qcf_neb.nextchain",
            [{"info_dict": service_state.nebinfo}],
            service_orm.compute_tag,
            service_orm.compute_priority,
            neb_orm.creator_user_id,
            session=session,
        )

        svc_dep = ServiceDependencyORM(
            record_id=ids[0],
            extras={},
        )

        service_orm.dependencies.append(svc_dep)

    def insert_full_qcportal_records_v1(
        self, session: Session, records: Sequence[NEBRecord], creator_user_id: Optional[int]
    ) -> List[NEBRecordORM]:
        ret = []

        neb_specs = [r.specification for r in records]

        meta, spec_ids = self.root_socket.records.neb.add_specifications(neb_specs, session=session)
        if not meta.success:
            raise RuntimeError("Aborted NEB insertion - could not add specifications: " + meta.error_string)

        for record, spec_id in zip(records, spec_ids):

            record_orm = NEBRecordORM(
                specification_id=spec_id,
                status=RecordStatusEnum.complete,
            )

            meta, init_chain_mol_ids = self.root_socket.molecules.add(record.initial_chain_, session=session)
            if not meta.success:
                raise RuntimeError("Aborted NEB insertion - could not add initial chain: " + meta.error_string)

            record_orm.initial_chain = [
                NEBInitialchainORM(molecule_id=mid, position=pos) for pos, mid in enumerate(init_chain_mol_ids)
            ]

            if record.singlepoint_records_:
                record_orm.singlepoints = []

                for chain_iteration, singlepoint_records in record.singlepoint_records_.items():
                    sp_ids = self.root_socket.records.insert_full_qcportal_records(
                        session, singlepoint_records, creator_user_id
                    )

                    for idx, sp_id in enumerate(sp_ids):
                        ns = NEBSinglepointsORM(
                            chain_iteration=chain_iteration,
                            position=idx,
                            singlepoint_id=sp_id,
                        )
                        record_orm.singlepoints.append(ns)

            if record.ts_hessian_:
                hessian_id = self.root_socket.records.insert_full_qcportal_records(
                    session, [record.ts_hessian_], creator_user_id
                )
                ns = NEBSinglepointsORM(
                    chain_iteration=max(record.singlepoint_records_.keys()) + 1,
                    position=0,
                    singlepoint_id=hessian_id[0],
                )
                record_orm.singlepoints.append(ns)

            if record.optimization_records_:
                record_orm.optimizations = []

                opt_info = list(record.optimization_records_.items())

                opt_records = [x[1] for x in opt_info]
                opt_ids = self.root_socket.records.insert_full_qcportal_records(session, opt_records, creator_user_id)

                for opt_record, opt_id, (k, _) in zip(opt_records, opt_ids, opt_info):
                    o_meta = record.optimizations_[k]
                    opt_orm = NEBOptimizationsORM(
                        optimization_id=opt_id,
                        position=o_meta.position,
                        ts=o_meta.ts,
                    )
                    record_orm.optimizations.append(opt_orm)

            ret.append(record_orm)

        return ret

    def add_specifications(
        self, neb_specs: Sequence[NEBSpecification], *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, List[int]]:
        """
        Adds specifications for NEB services to the database, returning their IDs.

        If an identical specification exists, then no insertion takes place and the id of the existing
        specification is returned.

        Specification IDs are returned in the same order as the input specifications

        Parameters
        ----------
        neb_specs
            Sequence of specifications to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and the IDs of the specifications.
        """

        to_add = []

        for neb_spec in neb_specs:
            kw_dict = neb_spec.keywords.dict()

            neb_spec_dict = {"program": neb_spec.program, "keywords": kw_dict, "protocols": {}}
            neb_spec_hash = hash_dict(neb_spec_dict)

            neb_spec_orm = NEBSpecificationORM(
                program=neb_spec.program,
                keywords=kw_dict,
                protocols=neb_spec_dict["protocols"],
                specification_hash=neb_spec_hash,
            )

            to_add.append(neb_spec_orm)

        with self.root_socket.optional_session(session, False) as session:

            qc_specs = [x.singlepoint_specification for x in neb_specs]
            meta, qc_spec_ids = self.root_socket.records.singlepoint.add_specifications(qc_specs, session=session)

            if not meta.success:
                return (
                    InsertMetadata(
                        error_description="Unable to add singlepoint specifications: " + meta.error_string,
                    ),
                    [],
                )

            opt_specs_lst = [
                (idx, x.optimization_specification)
                for idx, x in enumerate(neb_specs)
                if x.optimization_specification is not None
            ]

            opt_specs_map = {}
            if opt_specs_lst:
                opt_specs = [x[1] for x in opt_specs_lst]
                meta, opt_spec_ids = self.root_socket.records.optimization.add_specifications(
                    opt_specs, session=session
                )

                if not meta.success:
                    return (
                        InsertMetadata(
                            error_description="Unable to add optimization specifications: " + meta.error_string,
                        ),
                        [],
                    )

                opt_specs_map = {idx: opt_spec_id for (idx, _), opt_spec_id in zip(opt_specs_lst, opt_spec_ids)}

            for idx, (neb_spec_orm, qc_spec_id) in enumerate(zip(to_add, qc_spec_ids)):
                neb_spec_orm.singlepoint_specification_id = qc_spec_id
                neb_spec_orm.optimization_specification_id = opt_specs_map.get(idx, None)

            meta, ids = insert_general(
                session,
                to_add,
                (
                    NEBSpecificationORM.specification_hash,
                    NEBSpecificationORM.singlepoint_specification_id,
                    NEBSpecificationORM.optimization_specification_id,
                ),
                (NEBSpecificationORM.id,),
                use_unique=True,
            )

            return meta, [x[0] for x in ids]

    def add_specification(
        self, neb_spec: NEBSpecification, *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, Optional[int]]:
        """
        Adds a specification for an NEB service to the database, returning its id.

        If an identical specification exists, then no insertion takes place and the id of the existing
        specification is returned.

        Parameters
        ----------
        neb_spec
            Specification to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and the id of the specification.
        """

        meta, ids = self.add_specifications([neb_spec], session=session)

        if not ids:
            return meta, None

        return meta, ids[0]

    def get(
        self,
        record_ids: Sequence[int],
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
        missing_ok: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> List[Optional[Dict[str, Any]]]:
        options = []

        if include:
            if is_included("initial_chain", include, exclude, False) or is_included(
                "molecules", include, exclude, False
            ):
                options.append(selectinload(NEBRecordORM.initial_chain).joinedload(NEBInitialchainORM.molecule))
            if is_included("singlepoints", include, exclude, False):
                options.append(selectinload(NEBRecordORM.singlepoints))
            if is_included("optimizations", include, exclude, False):
                options.append(selectinload(NEBRecordORM.optimizations))

        with self.root_socket.optional_session(session, True) as session:
            return self.root_socket.records.get_base(
                orm_type=self.record_orm,
                record_ids=record_ids,
                include=include,
                exclude=exclude,
                missing_ok=missing_ok,
                additional_options=options,
                session=session,
            )

    def query(
        self,
        query_data: NEBQueryFilters,
        *,
        session: Optional[Session] = None,
    ) -> List[int]:
        and_query = []
        need_spspec_join = False
        need_initchain_join = False
        need_nebspec_join = False

        if query_data.qc_program is not None:
            and_query.append(QCSpecificationORM.program.in_(query_data.qc_program))
            need_spspec_join = True
        if query_data.qc_method is not None:
            and_query.append(QCSpecificationORM.method.in_(query_data.qc_method))
            need_spspec_join = True
        if query_data.qc_basis is not None:
            and_query.append(QCSpecificationORM.basis.in_(query_data.qc_basis))
            need_spspec_join = True
        if query_data.program is not None:
            and_query.append(NEBSpecificationORM.program.in_(query_data.program))
            need_nebspec_join = True
        if query_data.molecule_id is not None:
            and_query.append(NEBInitialchainORM.molecule_id.in_(query_data.molecule_id))
            need_initchain_join = True

        stmt = select(NEBRecordORM.id)

        if need_nebspec_join or need_spspec_join:
            stmt = stmt.join(NEBRecordORM.specification)

        if need_spspec_join:
            stmt = stmt.join(NEBSpecificationORM.singlepoint_specification)

        if need_initchain_join:
            stmt = stmt.join(
                NEBInitialchainORM,
                NEBInitialchainORM.neb_id == NEBRecordORM.id,
            )

        stmt = stmt.where(*and_query)

        return self.root_socket.records.query_base(
            stmt=stmt,
            orm_type=NEBRecordORM,
            query_data=query_data,
            session=session,
        )

    def add_internal(
        self,
        initial_chain_ids: Sequence[Sequence[int]],
        neb_spec_id: int,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user_id: Optional[int],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Internal function for adding new NEB calculations

        This function expects that the chains and specifications are already added to the
        database and that the ids are known.

        This checks if the calculations already exist in the database. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        Parameters
        ----------
        initial_chain_ids
            IDs of the chains to optimize. One record will be added per chain.
        neb_spec_id
            ID of the specification
        compute_tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        compute_priority
            The priority for the computation
        creator_user_id
            ID of the user who created the record
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of record ids. The ids will be in the
            order of the input chains
        """
        compute_tag = compute_tag.lower()

        with self.root_socket.optional_session(session, False) as session:
            # Lock for the entire transaction
            session.execute(select(func.pg_advisory_xact_lock(neb_insert_lock_id))).scalar()

            neb_ids = []
            inserted_idx = []
            existing_idx = []

            if find_existing:
                init_mol_cte = (
                    select(
                        NEBRecordORM.id,
                        NEBRecordORM.specification_id,
                        array_agg(
                            aggregate_order_by(
                                NEBInitialchainORM.molecule_id,
                                NEBInitialchainORM.position.asc(),
                            )
                        ).label("molecule_ids"),
                    )
                    .join(
                        NEBInitialchainORM,
                        NEBInitialchainORM.neb_id == NEBRecordORM.id,
                    )
                    .group_by(NEBRecordORM.id)
                    .cte()
                )

                for idx, mol_ids in enumerate(initial_chain_ids):
                    # does this exist?
                    stmt = select(init_mol_cte.c.id)
                    stmt = stmt.where(init_mol_cte.c.specification_id == neb_spec_id)
                    stmt = stmt.where(init_mol_cte.c.molecule_ids == mol_ids)
                    existing = session.execute(stmt).scalars().first()

                    if not existing:
                        neb_orm = NEBRecordORM(
                            is_service=True,
                            specification_id=neb_spec_id,
                            status=RecordStatusEnum.waiting,
                            creator_user_id=creator_user_id,
                        )

                        self.create_service(neb_orm, compute_tag, compute_priority, find_existing)

                        session.add(neb_orm)
                        session.flush()

                        for pos, mid in enumerate(mol_ids):
                            mid_orm = NEBInitialchainORM(molecule_id=mid, neb_id=neb_orm.id, position=pos)
                            session.add(mid_orm)

                        session.flush()

                        neb_ids.append(neb_orm.id)
                        inserted_idx.append(idx)
                    else:
                        neb_ids.append(existing)
                        existing_idx.append(idx)
            else:
                for idx, mol_ids in enumerate(initial_chain_ids):
                    neb_orm = NEBRecordORM(
                        is_service=True,
                        specification_id=neb_spec_id,
                        status=RecordStatusEnum.waiting,
                        creator_user_id=creator_user_id,
                    )

                    self.create_service(neb_orm, compute_tag, compute_priority, find_existing)

                    session.add(neb_orm)
                    session.flush()

                    for pos, mid in enumerate(mol_ids):
                        mid_orm = NEBInitialchainORM(molecule_id=mid, neb_id=neb_orm.id, position=pos)
                        session.add(mid_orm)

                    session.flush()

                    neb_ids.append(neb_orm.id)
                    inserted_idx.append(idx)

            meta = InsertMetadata(inserted_idx=inserted_idx, existing_idx=existing_idx)
            return meta, neb_ids

    def add(
        self,
        initial_chains: Sequence[Sequence[Union[int, Molecule]]],
        neb_spec: NEBSpecification,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Adds new neb calculations

        This checks if the calculations already exist in the database. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        If session is specified, changes are not committed to the database, but the session is flushed.

        Parameters
        ----------
        initial_chains
            Molecules to compute using the specification
        neb_spec
            Specification for the NEB calculations
        compute_tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        compute_priority
            The priority for the computation
        creator_user
            Name or ID of the user who created the record
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of record ids. The ids will be in the
            order of the input molecules
        """
        images = neb_spec.keywords.images
        with self.root_socket.optional_session(session, False) as session:
            creator_user_id = self.root_socket.users.get_optional_user_id(creator_user, session=session)

            # First, add the specification
            spec_meta, spec_id = self.add_specification(neb_spec, session=session)
            if not spec_meta.success:
                return (
                    InsertMetadata(
                        error_description="Aborted - could not add specification: " + spec_meta.error_string
                    ),
                    [],
                )

            # Now the molecules
            init_molecule_ids = []
            for init_chain in initial_chains:
                if len(init_chain) < images:
                    return (
                        InsertMetadata(
                            error_description="Aborted - number of images for NEB can not exceed the number of input frames:"
                            + spec_meta.error_string
                        ),
                        [],
                    )
                selected_chain = np.array(init_chain)[
                    np.array([int(round(i)) for i in np.linspace(0, len(init_chain) - 1, images)])
                ]

                mol_meta, molecule_ids = self.root_socket.molecules.add_mixed(selected_chain, session=session)
                if not mol_meta.success:
                    return (
                        InsertMetadata(
                            error_description="Aborted - could not add all molecules: " + mol_meta.error_string
                        ),
                        [],
                    )

                init_molecule_ids.append(molecule_ids)

            return self.add_internal(
                init_molecule_ids,
                spec_id,
                compute_tag,
                compute_priority,
                creator_user_id,
                find_existing,
                session=session,
            )

    def add_from_input(
        self,
        record_input: NEBInput,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertCountsMetadata, int]:

        assert isinstance(record_input, NEBInput)

        meta, ids = self.add(
            [record_input.initial_chain],
            record_input.specification,
            compute_tag,
            compute_priority,
            creator_user,
            find_existing,
        )

        return InsertCountsMetadata.from_insert_metadata(meta), ids[0]

    ####################################################
    # Some stuff to be retrieved for NEB records
    ####################################################

    def get_initial_molecules_ids(
        self,
        record_id: int,
        *,
        session: Optional[Session] = None,
    ) -> List[int]:
        options = [
            lazyload("*"),
            defer("*"),
            joinedload(NEBRecordORM.initial_chain).options(undefer(NEBInitialchainORM.molecule_id)),
        ]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(NEBRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")
            return [x.molecule_id for x in rec.initial_chain]

    def get_singlepoints(
        self,
        record_id: int,
        *,
        session: Optional[Session] = None,
    ) -> List[Dict[str, Any]]:
        options = [lazyload("*"), defer("*"), joinedload(NEBRecordORM.singlepoints).options(undefer("*"))]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(NEBRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")
            return [x.model_dict() for x in rec.singlepoints]

    def get_optimizations(
        self,
        record_id: int,
        *,
        session: Optional[Session] = None,
    ) -> Dict[str, Dict[str, Any]]:
        options = [lazyload("*"), defer("*"), joinedload(NEBRecordORM.optimizations).options(undefer("*"))]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(NEBRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")

            return rec.model_dict()["optimizations"]

    def get_neb_result(
        self,
        neb_id: int,
        *,
        session: Optional[Session] = None,
    ) -> Dict[str, Any]:
        stmt = (
            select(MoleculeORM)
            .join(SinglepointRecordORM)
            .join(NEBSinglepointsORM)
            .where(NEBSinglepointsORM.neb_id == neb_id)
            .order_by(
                NEBSinglepointsORM.chain_iteration.desc(),
                SinglepointRecordORM.properties["return_energy"].cast(TEXT).cast(DOUBLE_PRECISION).desc(),
            )
            .limit(1)
        )

        with self.root_socket.optional_session(session, True) as session:
            r = session.execute(stmt).scalar_one_or_none()
            if r is None:
                raise MissingDataError("The final guessed transition state can't be found")
            return r.model_dict()



from __future__ import annotations

from typing import TYPE_CHECKING, Tuple, Optional
from urllib.parse import urlparse

from flask import request, g, current_app, session
from flask_jwt_extended import (
    create_access_token,
    create_refresh_token,
)

from qcfractal import __version__ as qcfractal_version
from qcfractal.flask_app import storage_socket
from qcportal.auth import UserInfo
from qcportal.exceptions import AuthenticationFailure

if TYPE_CHECKING:
    from typing import Set

_all_endpoints: Set[str] = set()


def get_all_endpoints() -> Set[str]:
    """
    Get a list of all endpoints on the server

    These endpoints are the first three parts of the resource (ie,
    /api/v1/molecules, not /api/v1/molecules/bulkGet)
    """
    global _all_endpoints

    if not _all_endpoints:
        for url in current_app.url_map.iter_rules():
            endpoint = get_url_major_component(url.rule)
            # Don't add "static"
            if not endpoint.startswith("/static"):
                _all_endpoints.add(endpoint)

    return _all_endpoints


def get_url_major_component(url: str):
    """
    Obtains the major parts of a URL's components

    For example, /api/v1/molecule/a/b/c -> /api/v1/molecule
    """

    components = urlparse(url).path.split("/")
    resource = "/".join(components[:4])

    # Force leading slash, but only one
    return "/" + resource.lstrip("/")


def login_user() -> UserInfo:
    """
    Handle a login from flask

    This function authenticates the username/password sent to flask, and returns the JWT tokens.
    It handles the username/password being stored in json as well as form data.

    If get_refresh_token is True, then a refresh token is also return. Otherwise, None is returned
    for the refresh token.

    Returns
    -------
    :
        The access token and optionally the refresh token
    """
    try:
        if request.is_json:
            username = request.json["username"]
            password = request.json["password"]
        else:
            username = request.form["username"]
            password = request.form["password"]
    except Exception:
        current_app.logger.info("Invalid/malformed login request")
        raise AuthenticationFailure("Invalid/malformed login request")

    if username is None:
        current_app.logger.info("No username provided for login")
        raise AuthenticationFailure("No username provided for login")
    if password is None:
        current_app.logger.info(f"No password provided for login of user {username}")
        raise AuthenticationFailure("No password provided for login")

    try:
        user_info = storage_socket.users.authenticate(username, password)

        # Used for logging (in the after_request_func)
        g.user_id = user_info.id

        return user_info

    except AuthenticationFailure as e:
        current_app.logger.info(f"Authentication failed for user {username}: {str(e)}")
        raise


def login_user_session() -> UserInfo:
    # Raises exception on invalid username, password, etc
    # Submitted user/password are stored in the flask request object
    session.clear()
    user_info = login_user()
    session["user_id"] = str(user_info.id)

    return user_info


def logout_user_session():
    session.clear()


def access_token_from_user(user_info: UserInfo):
    """
    Creates a JWT access token from user/role information
    """
    return create_access_token(
        identity=str(user_info.id),
        additional_claims={
            "username": user_info.username,
            "role": user_info.role,
            "groups": user_info.groups,
        },
    )


def login_and_get_jwt(get_refresh_token: bool) -> Tuple[str, Optional[str]]:
    """
    Handle a login from flask

    This function authenticates the username/password sent to flask, and returns the JWT tokens.
    It handles the username/password being stored in json as well as form data.

    If get_refresh_token is True, then a refresh token is also return. Otherwise, None is returned
    for the refresh token.

    Returns
    -------
    :
        The access token and optionally the refresh token
    """

    # Will raise exceptions on invalid username/password
    user_info = login_user()
    access_token = access_token_from_user(user_info)

    if get_refresh_token:
        refresh_token = create_refresh_token(identity=str(user_info.id))
    else:
        refresh_token = None

    current_app.logger.info(f"Successful login for user {user_info.username}")
    return access_token, refresh_token


def get_public_server_information():
    qcf_cfg = current_app.config["QCFRACTAL_CONFIG"]

    # TODO - remove version limits after a while. They are there to support older clients
    public_info = {
        "name": qcf_cfg.name,
        "manager_heartbeat_frequency": qcf_cfg.heartbeat_frequency,
        "manager_heartbeat_frequency_jitter": qcf_cfg.heartbeat_frequency_jitter,
        "manager_heartbeat_max_missed": qcf_cfg.heartbeat_max_missed,
        "version": qcfractal_version,
        "api_limits": qcf_cfg.api_limits.model_dump(mode="json"),
        "client_version_lower_limit": "0.50",
        "client_version_upper_limit": "1.00",
        "manager_version_lower_limit": "0.50",
        "manager_version_upper_limit": "1.00",
        "motd": storage_socket.serverinfo.get_motd(),
    }

    return public_info



from typing import Dict

from flask import current_app, g

from qcfractal.flask_app import storage_socket
from qcfractal.flask_app.api_v1.blueprint import api_v1
from qcfractal.flask_app.wrap_route import wrap_global_route
from qcportal.base_models import ProjURLParameters
from qcportal.dataset_models import (
    DatasetAddBody,
    DatasetQueryModel,
    DatasetFetchRecordsBody,
    DatasetFetchEntryBody,
    DatasetFetchSpecificationBody,
    DatasetCreateViewBody,
    DatasetSubmitBody,
    DatasetDeleteStrBody,
    DatasetRecordModifyBody,
    DatasetRemoveRecordsBody,
    DatasetRecordRevertBody,
    DatasetModifyMetadata,
    DatasetQueryRecords,
    DatasetDeleteParams,
    DatasetModifyEntryBody,
    DatasetGetInternalJobParams,
    DatasetCloneBody,
    DatasetCopyFromBody,
)
from qcportal.exceptions import LimitExceededError


@api_v1.route("/datasets", methods=["GET"])
@wrap_global_route("datasets", "read")
def list_dataset_v1():
    return storage_socket.datasets.list()


@api_v1.route("/datasets/<int:dataset_id>", methods=["GET"])
@wrap_global_route("datasets", "read")
def get_general_dataset_v1(dataset_id: int, url_params: ProjURLParameters):
    with storage_socket.session_scope(True) as session:
        ds_type = storage_socket.datasets.lookup_type(dataset_id, session=session)
        ds_socket = storage_socket.datasets.get_socket(ds_type)

        r = ds_socket.get(dataset_id, url_params.include, url_params.exclude, session=session)

        # TODO - remove this eventually
        # Don't return attachments by default
        r.pop("attachments", None)

        return r


@api_v1.route("/datasets/query", methods=["POST"])
@wrap_global_route("datasets", "read")
def query_general_dataset_v1(body_data: DatasetQueryModel):
    with storage_socket.session_scope(True) as session:
        dataset_id = storage_socket.datasets.lookup_id(body_data.dataset_type, body_data.dataset_name, session=session)
        ds_type = storage_socket.datasets.lookup_type(dataset_id, session=session)
        ds_socket = storage_socket.datasets.get_socket(ds_type)
        return ds_socket.get(dataset_id, body_data.include, body_data.exclude, session=session)


@api_v1.route("/datasets/queryrecords", methods=["POST"])
@wrap_global_route("datasets", "read")
def query_dataset_records_v1(body_data: DatasetQueryRecords):
    return storage_socket.datasets.query_dataset_records(
        record_id=body_data.record_id, dataset_type=body_data.dataset_type
    )


@api_v1.route("/datasets/<int:dataset_id>", methods=["DELETE"])
@wrap_global_route("datasets", "delete")
def delete_dataset_v1(dataset_id: int, url_params: DatasetDeleteParams):
    storage_socket.datasets.delete(dataset_id, url_params.delete_records)


#################################################################
# COMMON FUNCTIONS
# These functions are common to all datasets
# Note that the inputs are all the same, but the returned dicts
# are different
#################################################################


########################
# Adding a dataset
########################
@api_v1.route("/datasets/<string:dataset_type>", methods=["POST"])
@wrap_global_route("datasets", "modify")
def add_dataset_v1(dataset_type: str, body_data: DatasetAddBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.add(
        name=body_data.name,
        description=body_data.description,
        tagline=body_data.tagline,
        tags=body_data.tags,
        provenance=body_data.provenance,
        default_compute_tag=body_data.default_compute_tag,
        default_compute_priority=body_data.default_compute_priority,
        extras=body_data.extras,
        creator_user=g.username,
        existing_ok=body_data.existing_ok,
    )


#########################
# Getting info
#########################
@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>", methods=["GET"])
@wrap_global_route("datasets", "read")
def get_dataset_v1(dataset_type: str, dataset_id: int, url_params: ProjURLParameters):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.get(
        dataset_id,
        url_params.include,
        url_params.exclude,
    )


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/status", methods=["GET"])
@wrap_global_route("datasets", "read")
def get_dataset_status_v1(dataset_type: str, dataset_id: int):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.status(dataset_id)


@api_v1.route("/datasets/<int:dataset_id>/status", methods=["GET"])
@wrap_global_route("datasets", "read")
def get_base_dataset_status_v1(dataset_id: int):
    return storage_socket.datasets.status(dataset_id)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/detailed_status", methods=["GET"])
@wrap_global_route("datasets", "read")
def get_dataset_detailed_status_v1(dataset_type: str, dataset_id: int):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.detailed_status(dataset_id)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/status_by_tag", methods=["GET"])
@wrap_global_route("datasets", "read")
def get_dataset_status_by_tag_v1(dataset_type: str, dataset_id: int):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.status_by_compute_tag(dataset_id)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/record_count", methods=["GET"])
@wrap_global_route("datasets", "read")
def get_dataset_record_count_v1(dataset_type: str, dataset_id: int):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.get_record_count(dataset_id)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/computed_properties", methods=["GET"])
@wrap_global_route("datasets", "read")
def get_dataset_computed_properties_v1(dataset_type: str, dataset_id: int):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.get_computed_properties(dataset_id)


#########################
# Modifying metadata
#########################
@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>", methods=["PATCH"])
@wrap_global_route("datasets", "modify")
def modify_dataset_metadata_v1(dataset_type: str, dataset_id: int, body_data: DatasetModifyMetadata):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.update_metadata(dataset_id, new_metadata=body_data)


#########################
# Views
#########################
@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/create_view", methods=["POST"])
@wrap_global_route("datasets", "modify")
def create_dataset_view_v1(dataset_type: str, dataset_id: int, body_data: DatasetCreateViewBody):
    return storage_socket.datasets.add_create_view_attachment_job(
        dataset_id,
        dataset_type,
        description=body_data.description,
        provenance=body_data.provenance,
        status=body_data.status,
        include=body_data.include,
        exclude=body_data.exclude,
        include_children=body_data.include_children,
    )


#########################
# Computation submission
#########################
@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/submit", methods=["POST"])
@wrap_global_route("datasets", "modify")
def submit_dataset_v1(dataset_type: str, dataset_id: int, body_data: DatasetSubmitBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.submit(
        dataset_id,
        entry_names=body_data.entry_names,
        specification_names=body_data.specification_names,
        compute_tag=body_data.compute_tag,
        compute_priority=body_data.compute_priority,
        creator_user=g.username,
        find_existing=body_data.find_existing,
    )


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/background_submit", methods=["POST"])
@wrap_global_route("datasets", "modify")
def background_submit_dataset_v1(dataset_type: str, dataset_id: int, body_data: DatasetSubmitBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.background_submit(
        dataset_id,
        entry_names=body_data.entry_names,
        specification_names=body_data.specification_names,
        compute_tag=body_data.compute_tag,
        compute_priority=body_data.compute_priority,
        creator_user=g.username,
        find_existing=body_data.find_existing,
    )


###################
# Specifications
###################
@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/specification_names", methods=["GET"])
@wrap_global_route("datasets", "read")
def fetch_dataset_specification_names_v1(dataset_type: str, dataset_id: int):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.fetch_specification_names(dataset_id)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/specifications", methods=["GET"])
@wrap_global_route("datasets", "read")
def fetch_all_dataset_specifications_v1(dataset_type: str, dataset_id: int):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.fetch_specifications(dataset_id)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/specifications/bulkFetch", methods=["POST"])
@wrap_global_route("datasets", "read")
def fetch_dataset_specifications_v1(dataset_type: str, dataset_id: int, body_data: DatasetFetchSpecificationBody):
    # use the entry limit I guess?
    limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_dataset_entries

    if len(body_data.names) > limit:
        raise LimitExceededError(f"Cannot get {len(body_data.names)} dataset specifications - limit is {limit}")

    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.fetch_specifications(
        dataset_id,
        specification_names=body_data.names,
        missing_ok=body_data.missing_ok,
    )


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/specifications/bulkDelete", methods=["POST"])
@wrap_global_route("datasets", "modify")
def delete_dataset_specifications_v1(dataset_type: str, dataset_id: int, body_data: DatasetDeleteStrBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.delete_specifications(dataset_id, body_data.names, body_data.delete_records)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/specifications", methods=["PATCH"])
@wrap_global_route("datasets", "modify")
def rename_dataset_specifications_v1(dataset_type: str, dataset_id: int, body_data: Dict[str, str]):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.rename_specifications(dataset_id, body_data)


###################
# Entries
###################
@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/entry_names", methods=["GET"])
@wrap_global_route("datasets", "read")
def fetch_dataset_entry_names_v1(dataset_type: str, dataset_id: int):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.fetch_entry_names(dataset_id)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/entries/bulkDelete", methods=["POST"])
@wrap_global_route("datasets", "modify")
def delete_dataset_entries_v1(dataset_type: str, dataset_id: int, body_data: DatasetDeleteStrBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.delete_entries(dataset_id, body_data.names, body_data.delete_records)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/entries/bulkFetch", methods=["POST"])
@wrap_global_route("datasets", "read")
def fetch_dataset_entries_v1(dataset_type: str, dataset_id: int, body_data: DatasetFetchEntryBody):
    limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_dataset_entries

    if len(body_data.names) > limit:
        raise LimitExceededError(f"Cannot get {len(body_data.names)} dataset entries - limit is {limit}")

    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.fetch_entries(
        dataset_id,
        entry_names=body_data.names,
        missing_ok=body_data.missing_ok,
    )


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/background_add_entries", methods=["POST"])
@wrap_global_route("datasets", "modify")
def background_add_entries_v1(dataset_type: str, dataset_id: int, body_data: DatasetSubmitBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.background_submit(
        dataset_id,
        entry_names=body_data.entry_names,
        specification_names=body_data.specification_names,
        compute_tag=body_data.compute_tag,
        compute_priority=body_data.compute_priority,
        creator_user=g.username,
        find_existing=body_data.find_existing,
    )


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/entries", methods=["PATCH"])
@wrap_global_route("datasets", "modify")
def rename_dataset_entries_v1(dataset_type: str, dataset_id: int, body_data: Dict[str, str]):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.rename_entries(dataset_id, body_data)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/entries/modify", methods=["PATCH"])
@wrap_global_route("datasets", "modify")
def modify_dataset_entries_v1(dataset_type: str, dataset_id: int, body_data: DatasetModifyEntryBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.modify_entries(
        dataset_id, body_data.attribute_map, body_data.comment_map, body_data.overwrite_attributes
    )


#########################
# Records
#########################
@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/records/bulkFetch", methods=["POST"])
@wrap_global_route("datasets", "read")
def fetch_dataset_records_v1(dataset_type: str, dataset_id: int, body_data: DatasetFetchRecordsBody):
    limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_records

    n_requested = len(body_data.entry_names) * len(body_data.specification_names)
    if n_requested > limit:
        raise LimitExceededError(f"Cannot get {n_requested} dataset records - limit is {limit}")

    ds_socket = storage_socket.datasets.get_socket(dataset_type)

    return ds_socket.fetch_records(
        dataset_id,
        entry_names=body_data.entry_names,
        specification_names=body_data.specification_names,
        status=body_data.status,
    )


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/records/bulkDelete", methods=["POST"])
@wrap_global_route("datasets", "modify")
def remove_dataset_records_v1(dataset_type: str, dataset_id: int, body_data: DatasetRemoveRecordsBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.remove_records(
        dataset_id,
        entry_names=body_data.entry_names,
        specification_names=body_data.specification_names,
        delete_records=body_data.delete_records,
    )


#########################
# Record modification
#########################
@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/records", methods=["PATCH"])
@wrap_global_route("datasets", "modify")
def modify_dataset_records_v1(dataset_type: str, dataset_id: int, body_data: DatasetRecordModifyBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.modify_records(
        dataset_id,
        g.username,
        body_data.entry_names,
        body_data.specification_names,
        body_data.status,
        body_data.compute_priority,
        body_data.compute_tag,
        body_data.comment,
    )


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/records/revert", methods=["POST"])
@wrap_global_route("datasets", "modify")
def revert_dataset_records_v1(dataset_type: str, dataset_id: int, body_data: DatasetRecordRevertBody):
    ds_socket = storage_socket.datasets.get_socket(dataset_type)
    return ds_socket.revert_records(
        dataset_id,
        body_data.revert_status,
        body_data.entry_names,
        body_data.specification_names,
    )


#################################
# Internal Jobs
#################################
@api_v1.route("/datasets/<int:dataset_id>/internal_jobs/<int:job_id>", methods=["GET"])
@wrap_global_route("datasets", "read")
def get_dataset_internal_job_v1(dataset_id: int, job_id: int):
    return storage_socket.datasets.get_internal_job(dataset_id, job_id)


@api_v1.route("/datasets/<int:dataset_id>/internal_jobs", methods=["GET"])
@wrap_global_route("datasets", "read")
def list_dataset_internal_jobs_v1(dataset_id: int, url_params: DatasetGetInternalJobParams):
    return storage_socket.datasets.list_internal_jobs(dataset_id, status=url_params.status)


#################################
# Attachments
#################################
@api_v1.route("/datasets/<int:dataset_id>/attachments", methods=["GET"])
@wrap_global_route("datasets", "read")
def fetch_dataset_attachments_v1(dataset_id: int):
    return storage_socket.datasets.get_attachments(dataset_id)


@api_v1.route("/datasets/<int:dataset_id>/attachments/<int:attachment_id>", methods=["DELETE"])
@wrap_global_route("datasets", "modify")
def delete_dataset_attachment_v1(dataset_id: int, attachment_id: int):
    return storage_socket.datasets.delete_attachment(dataset_id, attachment_id)


#################################
# Contributed Values
#################################
@api_v1.route("/datasets/<int:dataset_id>/contributed_values", methods=["GET"])
@wrap_global_route("datasets", "read")
def fetch_dataset_contributed_values_v1(dataset_id: int):
    return storage_socket.datasets.get_contributed_values(dataset_id)


#################################
# Cloning and copying
#################################
@api_v1.route("/datasets/clone", methods=["POST"])
@wrap_global_route("datasets", "modify")
def clone_dataset_v1(body_data: DatasetCloneBody):
    with storage_socket.session_scope(True) as session:
        ds_type = storage_socket.datasets.lookup_type(body_data.source_dataset_id, session=session)
        ds_socket = storage_socket.datasets.get_socket(ds_type)
        return ds_socket.clone(body_data.source_dataset_id, body_data.new_dataset_name, session=session)


@api_v1.route("/datasets/<string:dataset_type>/<int:dataset_id>/copy_from", methods=["POST"])
@wrap_global_route("datasets", "modify")
def copy_from_dataset_v1(dataset_type: str, dataset_id: int, body_data: DatasetCopyFromBody):
    # the dataset_id in the URI is the destination dataset
    # (ie, the user has a dataset, then copies FROM another dataset
    with storage_socket.session_scope(True) as session:
        ds_socket = storage_socket.datasets.get_socket(dataset_type)
        return ds_socket.copy_from(
            body_data.source_dataset_id,
            dataset_id,
            body_data.entry_names,
            body_data.specification_names,
            body_data.copy_entries,
            body_data.copy_specifications,
            body_data.copy_records,
            session=session,
        )



from __future__ import annotations

import pytest

from qcportal import PortalClient
from qcportal.managers import ManagerName


@pytest.fixture(scope="module")
def queryable_managers_client(session_snowflake):
    for cluster_i in range(4):
        for host_i in range(2):
            for uuid_i in range(3):
                mname = ManagerName(
                    cluster=f"test_cluster_{cluster_i}",
                    hostname=f"test_host_{host_i}",
                    uuid=f"1234-5678-1234-567{uuid_i}",
                )

                mclient = session_snowflake.manager_client(mname)
                mclient.activate(
                    manager_version="v2.0",
                    programs={"qcengine": ["unknown"], "qcprog": ["unknown"], "qcprog2": ["v3.0"]},
                    compute_tags=[f"tag_{cluster_i}", "tag2"],
                )

                if uuid_i == 0:
                    mclient.deactivate(1, 1, 1.0, 1.0)

    yield session_snowflake.client()
    session_snowflake.reset()


def test_manager_client_query(queryable_managers_client: PortalClient):
    query_res = queryable_managers_client.query_managers(status="active")
    query_res_l = list(query_res)
    assert len(query_res_l) == 16

    query_res = queryable_managers_client.query_managers(hostname="test_host_1")
    query_res_l = list(query_res)
    assert len(query_res_l) == 12

    query_res = queryable_managers_client.query_managers(hostname="test_host_1", status="inactive")
    query_res_l = list(query_res)
    assert len(query_res_l) == 4

    query_res = queryable_managers_client.query_managers(cluster="test_cluster_2", status="active")
    query_res_l = list(query_res)
    assert len(query_res_l) == 4

    query_res = queryable_managers_client.query_managers(
        name=["test_cluster_2-test_host_1-1234-5678-1234-5672", "test_cluster_1-test_host_0-1234-5678-1234-5671"]
    )
    managers = list(query_res)
    assert len(managers) == 2

    query_res = queryable_managers_client.query_managers(manager_id=[managers[0].id, managers[1].id])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2


def test_manager_client_query_empty_iter(queryable_managers_client: PortalClient):
    query_res = queryable_managers_client.query_managers()
    assert len(query_res._current_batch) < queryable_managers_client.api_limits["get_managers"]

    managers = list(query_res)
    assert len(managers) == 24


def test_manager_client_query_limit(queryable_managers_client: PortalClient):
    query_res = queryable_managers_client.query_managers(limit=19)
    query_res_l = list(query_res)
    assert len(query_res_l) == 19
    assert len(query_res._current_batch) < queryable_managers_client.api_limits["get_managers"]

    query_res = queryable_managers_client.query_managers(hostname="test_host_1", limit=7)
    query_res_l = list(query_res)
    assert len(query_res_l) == 7


def test_manager_client_query_active(queryable_managers_client: PortalClient):
    mnames = queryable_managers_client.query_active_managers(compute_tag="tag2", programs={"qcprog": ["unknown"]})
    assert len(mnames) == 16

    mnames = queryable_managers_client.query_active_managers(
        compute_tag="tag_3", programs={"qcprog": ["unknown"], "qcprog2": ["v3.0"]}
    )
    assert len(mnames) == 4

    mnames = queryable_managers_client.query_active_managers(
        compute_tag=["tag_2", "othertag"], programs={"qcprog": ["unknown"], "qcprog2": ["v3.0"]}
    )
    assert len(mnames) == 4

    mnames = queryable_managers_client.query_active_managers(compute_tag="tag_3", programs={"qcprog9": ["unknown"]})
    assert len(mnames) == 0

    mnames = queryable_managers_client.query_active_managers(compute_tag="notag", programs={"qcprog": ["unknown"]})
    assert len(mnames) == 0

    mnames = queryable_managers_client.query_active_managers(compute_tag="tag2", programs={"qcprog9": ["unknown"]})
    assert len(mnames) == 0

    mnames = queryable_managers_client.query_active_managers(compute_tag="*", programs={"qcprog": ["unknown"]})
    assert len(mnames) == 16



from __future__ import annotations

import importlib
import logging
import logging.handlers
import os
import queue
import secrets
import shutil
import tempfile
import threading
import time
import weakref
from queue import Empty  # Just for exception handling
from typing import TYPE_CHECKING

import requests

from qcportal import PortalClient
from qcportal.client_base import AllowedConnectionExceptions
from qcportal.record_models import RecordStatusEnum
from qcportal.utils import update_nested_dict
from .config import FractalConfig, DatabaseConfig
from .port_util import find_open_port
from .postgres_harness import create_snowflake_postgres
from .process_targets import QCFJobRunnerThread, QCFAPIThread

if importlib.util.find_spec("qcfractalcompute") is None:
    raise RuntimeError("qcfractalcompute is not installed. Snowflake is useless without it")

from qcfractalcompute.config import FractalComputeConfig, FractalServerSettings, LocalExecutorConfig
from qcfractalcompute.process_targets import QCFComputeThread

if TYPE_CHECKING:
    from typing import Dict, Any, Sequence, Optional, Set


class FractalSnowflake:
    def __init__(
        self,
        start: bool = True,
        compute_workers: int = 2,
        database_config: Optional[DatabaseConfig] = None,
        extra_config: Optional[Dict[str, Any]] = None,
        *,
        host: str = "localhost",
        tmpdir_parent: Optional[str] = None,
    ):
        """A temporary, self-contained server

        A snowflake contains the server and compute manager, and can be used to test
        QCFractal/QCPortal, or to experiment.

        All data is lost when the server is shutdown.

        This can also be used as a context manager (`with FractalSnowflake(...) as s:`)
        """

        self._logger = logging.getLogger("fractal_snowflake")

        # Create a temporary directory for everything
        # Don't use TemporaryDirectory with autodelete - we will cleanup in the snowflake finalizer
        # If allowed to do it automatically, it may be cleaned up before the finalizer runs, and
        # the pg_harness won't like that (since the db files are stored there)
        # tempfile.TemporaryDirectory doesn't have a "delete" parameter in older versions of python.
        # So let's just use mkdtemp and delete it manually.
        # mktemp will return relative paths in some versions of python with some parameters of `dir`,
        # but I always feel more comfortable with absolute paths.
        self._tmpdir = os.path.abspath(tempfile.mkdtemp(dir=tmpdir_parent))

        # also parsl run directory and scratch directories
        parsl_run_dir = os.path.join(self._tmpdir, "parsl_run_dir")
        compute_scratch_dir = os.path.join(self._tmpdir, "compute_scratch_dir")
        os.makedirs(parsl_run_dir, exist_ok=True)
        os.makedirs(compute_scratch_dir, exist_ok=True)

        self._logger.info(f"Using temporary directory: {self._tmpdir}")

        if database_config is None:
            # db and socket are subdirs of the base temporary directory
            db_dir = os.path.join(self._tmpdir, "db")
            self._pg_harness = create_snowflake_postgres(host, db_dir)
            self._pg_harness.create_database(True)
            db_config = self._pg_harness.config
            self._own_pg_harness = True
        else:
            self._own_pg_harness = False  # self._pg_harness should be handled by derived class
            db_config = database_config

        api_port = find_open_port(host)
        self._fractal_uri = f"http://{host}:{api_port}"

        # Create a configuration for QCFractal
        # Assign the log level for subprocesses. Use the same level as what is assigned for this object
        loglevel = self._logger.getEffectiveLevel()

        qcf_cfg: Dict[str, Any] = {}
        qcf_cfg["base_folder"] = self._tmpdir
        qcf_cfg["loglevel"] = logging.getLevelName(loglevel)
        qcf_cfg["database"] = db_config.model_dump()
        qcf_cfg["enable_security"] = False
        qcf_cfg["hide_internal_errors"] = False
        qcf_cfg["service_frequency"] = 10
        qcf_cfg["heartbeat_frequency"] = 5
        qcf_cfg["heartbeat_frequency_jitter"] = 0.0
        qcf_cfg["heartbeat_max_missed"] = 3
        qcf_cfg["api"] = {
            "host": host,
            "port": api_port,
            "secret_key": secrets.token_urlsafe(32),
            "jwt_secret_key": secrets.token_urlsafe(32),
        }

        # Add in any options passed to this Snowflake
        if extra_config is not None:
            update_nested_dict(qcf_cfg, extra_config)

        self._qcf_config = FractalConfig(**qcf_cfg)

        ######################################
        # Set up the various components      #
        ######################################

        # For Flask
        self._finished_queue = queue.Queue()
        self._all_completed: Set[int] = set()
        self._api_thread = None
        self._api_initialized = threading.Event()

        # For the compute manager
        uri = f"http://{self._qcf_config.api.host}:{self._qcf_config.api.port}"
        self._compute_config = FractalComputeConfig(
            base_folder=self._tmpdir,
            loglevel=logging.getLevelName(loglevel),
            parsl_run_dir=parsl_run_dir,
            cluster="snowflake_compute",
            update_frequency=5,
            update_frequency_jitter=0.0,
            server=FractalServerSettings(
                fractal_uri=uri,
                verify=False,
            ),
            executors={
                "local": LocalExecutorConfig(
                    cores_per_worker=1,
                    memory_per_worker=1,
                    max_workers=compute_workers,
                    compute_tags=["*"],
                    scratch_directory=compute_scratch_dir,
                )
            },
        )
        self._compute_enabled = compute_workers > 0
        self._compute_thread: Optional[QCFComputeThread] = None
        self._compute_initialized = threading.Event()

        # Job runner
        self._job_runner_thread = None
        self._job_runner_initialized = threading.Event()

        # This is updated when starting components
        self._finalizer = None

        # Update now because of the logging thread & pg_harness
        self._update_finalizer()

        if start:
            self.start(wait=True)

    def _update_finalizer(self):
        if self._finalizer is not None:
            self._finalizer.detach()

        self._finalizer = weakref.finalize(
            self,
            self._stop,
            self._compute_thread,
            self._api_thread,
            self._job_runner_thread,
            self._pg_harness if self._own_pg_harness else None,
            self._tmpdir,
        )

    def _start_api(self, wait: bool = True):
        if self._api_thread is None:
            self._api_initialized.clear()
            self._api_thread = QCFAPIThread(self._qcf_config, self._finished_queue)
            self._api_thread.start(self._api_initialized)
            self._update_finalizer()

        if wait:
            self._api_initialized.wait()

    def _stop_api(self):
        if self._api_thread is not None:
            self._api_thread.stop()
            self._api_thread = None
            self._api_initialized.clear()
            self._update_finalizer()

    def _start_compute(self, wait: bool = True):
        if not self._compute_enabled:
            return

        if self._compute_thread is None:
            self._compute_initialized.clear()
            self._compute_thread = QCFComputeThread(self._compute_config)
            self._compute_thread.start(self._compute_initialized)
            self._update_finalizer()

        if wait:
            self._compute_initialized.wait()

    def _stop_compute(self):
        if self._compute_thread is not None:
            self._compute_thread.stop()
            self._compute_thread = None
            self._compute_initialized.clear()
            self._update_finalizer()

    def _start_job_runner(self, wait: bool = True):
        if self._job_runner_thread is None:
            self._job_runner_initialized.clear()
            self._job_runner_thread = QCFJobRunnerThread(self._qcf_config, self._finished_queue)
            self._job_runner_thread.start(self._job_runner_initialized)
            self._update_finalizer()

        if wait:
            self._job_runner_initialized.wait()

    def _stop_job_runner(self):
        if self._job_runner_thread is not None:
            self._job_runner_thread.stop()
            self._job_runner_thread = None
            self._job_runner_initialized.clear()
            self._update_finalizer()

    @classmethod
    def _stop(
        cls,
        compute_thread,
        api_thread,
        job_runner_thread,
        pg_harness,
        tmpdir,
    ):
        ####################################################################################
        # This is written as a class method so that it can be called by a weakref finalizer
        ####################################################################################

        # Stop these in a particular order
        # First the compute, since it will communicate its demise to the api server
        # Flask must be last. Then shutdown the database
        # (Note about the db - this function should only be called from the finalizer. Stopping/starting the db
        #  should NOT call this function, which allows data to persist between states of a particular
        #  snowflake object)

        if compute_thread is not None:
            compute_thread.stop()

        if job_runner_thread is not None:
            job_runner_thread.stop()

        if api_thread is not None:
            api_thread.stop()

        if pg_harness is not None:
            pg_harness.shutdown()

        if tmpdir and os.path.exists(tmpdir):
            shutil.rmtree(tmpdir)

    def wait_for_api(self):
        """
        Wait for the flask/api server to come up and then exit

        If it does not come up after some time, an exception will be raised
        """

        # Seems there still may be a small time after the event is triggered and before
        # it can handle requests
        # Can't use ping that is part of the client - we haven't instantiated one yet
        host = self._qcf_config.api.host
        port = self._qcf_config.api.port
        uri = f"http://{host}:{port}/api/v1/ping"

        max_iter = 100
        iter = 0
        while True:
            try:
                r = requests.get(uri)
                if r.status_code != 200:
                    raise RuntimeError("Error pinging snowflake fractal server: ", r.text)
                break

            except AllowedConnectionExceptions:
                time.sleep(0.1)
                iter += 1
                if iter >= max_iter:
                    raise

    def start(self, wait: bool = True) -> None:
        """
        Starts all the components of the snowflake
        """

        # Start all in parallel, then wait on all independently
        self._start_api(wait=False)
        self._start_compute(wait=False)
        self._start_job_runner(wait=False)

        if wait:
            if self._compute_thread is not None:
                self._compute_initialized.wait()

            if self._job_runner_thread is not None:
                self._job_runner_initialized.wait()

            if self._api_thread is not None:
                self._api_initialized.wait()
                self.wait_for_api()

    def stop(self):
        """
        Stops all components of the snowflake
        """

        self._stop_api()
        self._stop_compute()
        self._stop_job_runner()

    def get_uri(self) -> str:
        """
        Obtain the URI/address of the REST interface of this server

        Returns
        -------
        :
            Address/URI of the rest interface (ie, 'http://127.0.0.1:1234')
        """

        return f"http://{self._qcf_config.api.host}:{self._qcf_config.api.port}"

    def await_results(self, ids: Optional[Sequence[int]] = None, timeout: Optional[float] = None) -> bool:
        """
        Wait for computations to complete

        This function will block until the specified computations are complete (either success for failure).
        If timeout is given, that is the maximum amount of time to wait for a result. This timer is reset
        after each completed result.

        Parameters
        ----------
        ids
            Result/Procedure IDs to wait for. If not specified, all currently incomplete tasks
            will be waited for.

        timeout
            Maximum time to wait for a single result.


        Returns
        -------
        :
            True if all the results were received, False if timeout has elapsed without receiving a completed computation
        """
        logger = logging.getLogger(__name__)

        if ids is None:
            c = self.client()
            query_iter = c.query_records(status=[RecordStatusEnum.waiting, RecordStatusEnum.running])
            ids = [x.id for x in query_iter]

        # Remove any we have already marked as completed
        remaining_ids = set(ids) - self._all_completed

        if len(remaining_ids) == 0:
            logger.debug("All tasks are already finished")
            return True

        logger.debug("Waiting for ids: " + str(remaining_ids))

        while len(remaining_ids) > 0:
            # The queue stores a tuple of (id, status)
            try:
                finished_id, status = self._finished_queue.get(True, timeout)
            except Empty:
                logger.debug(f"Not all records finished in {timeout} seconds")
                return False

            logger.debug(f"Record finished: id={finished_id}, status={status}")

            # Add it to the list of all completed results we have seen
            self._all_completed.add(finished_id)

            # We may not be watching for this id, but if we are, remove it from the list
            # we are watching
            if finished_id in remaining_ids:
                remaining_ids.remove(finished_id)
                remaining_str = "None" if len(remaining_ids) == 0 else str(remaining_ids)
                logger.debug(f"Removed id={finished_id}. Remaining ids: {remaining_str}")

        return True

    def client(self) -> PortalClient:
        """
        Obtain a PortalClient connected to this server
        """

        # Shorten the timeout parameter - should be pretty quick in a snowflake
        c = PortalClient(self.get_uri())
        c._timeout = 2
        return c

    def dump_database(self, filepath: str) -> None:
        """
        Dumps the database to a file

        Parameters
        ----------
        filepath
            Path to the output file to create
        """

        self._pg_harness.backup_database(filepath)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.stop()



"""
Tests the general record socket
"""

from __future__ import annotations

import pytest

from qcarchivetesting import test_users
from qcarchivetesting.testing_classes import QCATestingSnowflake
from qcfractal.components.optimization.testing_helpers import (
    run_procedure_data as run_opt_procedure_data,
    submit_procedure_data as submit_opt_procedure_data,
)
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.components.singlepoint.testing_helpers import (
    run_procedure_data as run_sp_procedure_data,
    submit_procedure_data as submit_sp_procedure_data,
)
from qcfractal.components.testing_helpers import populate_records_status
from qcfractal.components.torsiondrive.testing_helpers import submit_procedure_data as submit_td_procedure_data
from qcportal import PortalRequestError
from qcportal.molecules import Molecule
from qcportal.record_models import PriorityEnum, RecordStatusEnum
from qcportal.utils import now_at_utc


def test_record_client_get(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    id1 = run_sp_procedure_data(storage_socket, activated_manager_name, "sp_psi4_benzene_energy_1")
    id2, _ = submit_opt_procedure_data(storage_socket, "opt_psi4_benzene")
    all_id = [id1, id2]

    r = snowflake_client.get_records(all_id)
    assert len(r)
    assert all_id == [x.id for x in r]
    assert [x.task is None for x in r]
    assert r[0].compute_history_ is None
    assert r[1].compute_history_ is None

    assert [x.task is None for x in r]

    r = snowflake_client.get_records(all_id, include=["compute_history", "task"])
    assert len(r[0].compute_history_) == 1
    assert len(r[1].compute_history_) == 0
    assert r[0].task is None
    assert r[1].task is not None


def test_record_client_get_missing(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    id1, _ = submit_sp_procedure_data(storage_socket, "sp_psi4_benzene_energy_1")
    id2, _ = submit_opt_procedure_data(storage_socket, "opt_psi4_benzene")

    all_id = [id1, id2]

    with pytest.raises(PortalRequestError, match=r"Could not find all requested"):
        snowflake_client.get_records([all_id[0], 9999, all_id[1]])

    r = snowflake_client.get_records([all_id[0], 9999, all_id[1]], missing_ok=True)
    assert r[1] is None
    assert r[0].id == all_id[0]
    assert r[2].id == all_id[1]


def test_record_client_get_empty(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    submit_sp_procedure_data(storage_socket, "sp_psi4_benzene_energy_1")
    submit_opt_procedure_data(storage_socket, "opt_psi4_benzene")

    r = snowflake_client.get_records([])
    assert r == []


def test_record_client_query_parents_children(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    id1 = run_opt_procedure_data(storage_socket, activated_manager_name, "opt_psi4_benzene")

    opt_rec = snowflake_client.get_optimizations(id1, include=["trajectory"])
    assert opt_rec.status == RecordStatusEnum.complete

    traj_id = [x.id for x in opt_rec.trajectory]
    assert len(traj_id) > 0

    # Query for records containing a record as a parent
    query_res = snowflake_client.query_records(parent_id=opt_rec.id)
    assert {x.id for x in query_res} == set(traj_id)

    # Query for records containing a record as a child
    query_res = snowflake_client.query_records(child_id=traj_id[0])
    query_res_l = list(query_res)
    assert len(query_res_l) == 1
    assert query_res_l[0].id == opt_rec.id


def test_record_client_add_comment(secure_snowflake: QCATestingSnowflake):
    storage_socket = secure_snowflake.get_storage_socket()

    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    id1, _ = submit_sp_procedure_data(storage_socket, "sp_psi4_benzene_energy_1")
    id2, _ = submit_sp_procedure_data(storage_socket, "sp_psi4_benzene_energy_2")
    id3, _ = submit_sp_procedure_data(storage_socket, "sp_psi4_benzene_energy_3")
    id4, _ = submit_opt_procedure_data(storage_socket, "opt_psi4_benzene")
    all_id = [id1, id2, id3, id4]

    # comments not retrieved by default
    rec = client.get_records(all_id)
    for r in rec:
        assert r.comments_ is None

    rec = client.get_records(all_id, include=["comments"])
    for r in rec:
        assert r.comments_ == []

    time_0 = now_at_utc()
    meta = client.add_comment([all_id[1], all_id[3]], comment="This is a test comment")
    time_1 = now_at_utc()
    assert meta.success
    assert meta.n_updated == 2
    assert meta.updated_idx == [0, 1]

    meta = client.add_comment([all_id[2], all_id[3]], comment="This is another test comment")
    time_2 = now_at_utc()
    assert meta.success
    assert meta.n_updated == 2
    assert meta.updated_idx == [0, 1]

    rec = client.get_records(all_id, include=["comments"])
    assert rec[0].comments == []
    assert len(rec[1].comments) == 1
    assert len(rec[2].comments) == 1
    assert len(rec[3].comments) == 2

    assert time_0 < rec[1].comments[0].timestamp < time_1
    assert time_1 < rec[2].comments[0].timestamp < time_2
    assert time_0 < rec[3].comments[0].timestamp < time_1
    assert time_1 < rec[3].comments[1].timestamp < time_2
    assert rec[1].comments[0].username == "admin_user"
    assert rec[3].comments[0].username == "admin_user"
    assert rec[2].comments[0].username == "admin_user"
    assert rec[3].comments[1].username == "admin_user"

    assert rec[1].comments[0].comment == "This is a test comment"
    assert rec[3].comments[0].comment == "This is a test comment"
    assert rec[2].comments[0].comment == "This is another test comment"
    assert rec[3].comments[1].comment == "This is another test comment"


def test_record_client_add_comment_nouser(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    id1, _ = submit_sp_procedure_data(storage_socket, "sp_psi4_benzene_energy_1")
    id2, _ = submit_sp_procedure_data(storage_socket, "sp_psi4_benzene_energy_2")
    id3, _ = submit_sp_procedure_data(storage_socket, "sp_psi4_benzene_energy_3")
    id4, _ = submit_opt_procedure_data(storage_socket, "opt_psi4_benzene")
    all_id = [id1, id2, id3, id4]

    time_0 = now_at_utc()
    meta = snowflake_client.add_comment([all_id[1], all_id[3]], comment="This is a test comment")
    time_1 = now_at_utc()
    assert meta.success
    assert meta.n_updated == 2
    assert meta.updated_idx == [0, 1]

    rec = snowflake_client.get_records(all_id, include=["comments"])
    assert len(rec[1].comments) == 1
    assert len(rec[3].comments) == 1

    assert time_0 < rec[1].comments[0].timestamp < time_1
    assert time_0 < rec[3].comments[0].timestamp < time_1
    assert rec[1].comments[0].username is None
    assert rec[3].comments[0].username is None

    assert rec[1].comments[0].comment == "This is a test comment"
    assert rec[3].comments[0].comment == "This is a test comment"


def test_record_client_add_comment_badid(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    id1, _ = submit_sp_procedure_data(storage_socket, "sp_psi4_benzene_energy_1")

    meta = snowflake_client.add_comment([id1, 9999], comment="test")
    assert not meta.success
    assert meta.n_updated == 1
    assert meta.n_errors == 1
    assert meta.updated_idx == [0]
    assert meta.error_idx == [1]
    assert "does not exist" in meta.errors[0][1]


def test_record_client_modify(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    all_id = populate_records_status(storage_socket)

    time_0 = now_at_utc()

    # record 2 is complete - can't change
    meta = snowflake_client.modify_records([all_id[0], all_id[1]], new_compute_tag="new_Tag")
    assert meta.n_updated == 1
    assert meta.updated_idx == [0]
    assert meta.error_idx == [1]

    # one of these records in cancelled
    meta = snowflake_client.modify_records([all_id[3], all_id[4]], new_compute_priority=PriorityEnum.low)
    assert meta.n_updated == 1

    rec = snowflake_client.get_records(all_id, include=["task"])

    # created_on and modified_on hasn't changed
    for r in rec:
        assert r.created_on < time_0
        assert r.modified_on < time_0

    # 0 - waiting
    assert rec[0].task.compute_tag == "new_tag"
    assert rec[0].task.compute_priority == PriorityEnum.normal

    # 1 - completed
    assert rec[1].task is None

    # 2 - running - not changed
    assert rec[2].task.compute_tag == "tag2"
    assert rec[2].task.compute_priority == PriorityEnum.high

    # 3 - error
    assert rec[3].task.compute_tag == "tag3"
    assert rec[3].task.compute_priority == PriorityEnum.low

    # 4/5/6 - cancelled/deleted/invalid
    assert rec[4].task is None
    assert rec[5].task is None
    assert rec[6].task is None

    rec = snowflake_client.get_records(all_id, include=["task"])

    # created_on and modified_on hasn't changed
    for r in rec:
        assert r.created_on < time_0
        assert r.modified_on < time_0

    assert rec[1].task is None
    assert rec[2].task.compute_priority == PriorityEnum.high
    assert rec[4].task is None
    assert rec[5].task is None
    assert rec[6].task is None


def test_record_client_modify_service(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    rec_id, _ = submit_td_procedure_data(storage_socket, "td_H2O2_mopac_pm6", "test_tag", PriorityEnum.high)

    with storage_socket.session_scope() as s:
        storage_socket.services.iterate_services(s)

        svc_id = s.get(BaseRecordORM, rec_id).service.id
        storage_socket.services._iterate_service(s, svc_id)

    with storage_socket.session_scope() as session:
        rec = session.get(BaseRecordORM, rec_id)
        tasks = [x.record.task for x in rec.service.dependencies]

        assert len(tasks) > 0
        assert all(x.compute_tag == "test_tag" for x in tasks)
        assert all(x.compute_priority == PriorityEnum.high for x in tasks)

    # Modify service priority and tag
    meta = snowflake_client.modify_records(rec_id, new_compute_tag="new_tag", new_compute_priority=PriorityEnum.low)
    assert meta.n_updated == 1
    assert meta.n_children_updated > 0

    rec = snowflake_client.get_records(rec_id)
    assert rec.service.compute_tag == "new_tag"
    assert rec.service.compute_priority == PriorityEnum.low

    # Also changed all the dependencies
    assert len(rec.service.dependencies) > 0
    for opt in rec.service.dependencies:
        r = snowflake_client.get_records(opt.record_id, include=["task"])
        assert r.task.compute_tag == "new_tag"
        assert r.task.compute_priority == PriorityEnum.low


def test_record_client_query_owner(secure_snowflake: QCATestingSnowflake):
    submit_client = secure_snowflake.client("submit_user", test_users["submit_user"]["pw"])
    admin_client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    submit_uid = submit_client.get_user().id

    m = Molecule(
        symbols=["h"],
        geometry=[0, 0, 0],
    )

    _, ids_1 = submit_client.add_singlepoints(m, "prog1", "energy", "b3lyp", "sto-3g", {})
    _, ids_2 = submit_client.add_singlepoints(m, "prog2", "energy", "b3lyp", "sto-3g", {})

    _, ids_3 = admin_client.add_singlepoints(m, "prog3", "energy", "b3lyp", "sto-3g", {})
    _, ids_4 = admin_client.add_singlepoints(m, "prog4", "energy", "b3lyp", "sto-3g", {})

    query_res = admin_client.query_records(creator_user="submit_user")
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    query_res = admin_client.query_records(creator_user=[submit_uid])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2
